from pygments.style import Style
from pygments.token import Keyword, Name, Comment, String, Error, \
     Number, Operator, Generic, Punctuation


# Style for Pygments source highlighter
class ApydiaDefaultStyle(Style):
    default_style = ""
    styles = {
        Comment:        "italic #999",
        Keyword:        "#00C",
        Name:           "#000",
        Name.Function:  "#F80",
        Name.Class:     "#F80",
        String:         "bg:#F8F8F8 #444",
        Punctuation:    "#444",
        Operator:       "#F80",
        Error:          "bg:#FFFFFF border:#F04 #802"
    }




# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: __init__.py 168 2008-04-10 14:12:10Z albrecht.andi $

"""Database backends"""

__doc_all__ = ["schema",]

import gobject
import gtk

import sys
import time

from cf.utils import Emit

TRANSACTION_IDLE = 1 << 1
TRANSACTION_COMMIT_ENABLED = 1 << 2
TRANSACTION_ROLLBACK_ENABLED = 1 << 3


class DBError(StandardError):
    """Base class for database errors"""
    
    def __init__(self, msg):
        self._msg = msg
    def __str__(self):
        return self._msg

class DBConnectError(DBError):
    """Errors on opening a connection"""
    
class DBConnection(gobject.GObject):
    
    __gsignals__ = {
        "closed" : (gobject.SIGNAL_RUN_LAST,
                    gobject.TYPE_NONE,
                    tuple()),
        "notice" : (gobject.SIGNAL_RUN_LAST,
                    gobject.TYPE_NONE,
                    (str,)),
    }
    
    __gproperties__ = {
        "transaction-state" : (gobject.TYPE_PYOBJECT,
                            "Transaction flag", "Transaction flag",
                            gobject.PARAM_READWRITE),
    }
    
    def __init__(self, provider, app):
        self.app = app
        self.provider = provider
        self.datasource_info = None
        self.conn_number = None
        self.threadsafety = 0
        self.coding_hint = "utf-8"
        self._transaction_state = TRANSACTION_IDLE
        self.__gobject_init__()
        
    def do_set_property(self, property, value):
        if property.name == "transaction-state":
            self._transaction_state = value
        else:
            raise AttributeError, "unknown property %r" % property.name
        
    def do_get_property(self, property):
        if property.name == "transaction-state":
            return self._transaction_state
        else:
            raise AttributeError, "unknown property %r" % property.name
        
    def get_label(self):
        return self.datasource_info.get_label() + " #%s" % self.conn_number
        
    def close(self):
        self.emit("closed")
        
    def cursor(self):
        raise NotImplementedError
    
    def update_transaction_status(self):
        pass
    
    def get_server_info(self):
        return None
    
    def commit(self):
        raise NotImplementedError
    
    def rollback(self):
        raise NotImplementedError
    
    def explain(self, statement):
        return []
    
class DBCursor(gobject.GObject):
    
    def __init__(self, connection):
        self.connection = connection
        self.__gobject_init__()
        
    def execute(self, query):
        raise NotImplementedError
    
    def get_messages(self):
        return []
    
    def close(self):
        raise NotImplementedError
    
    
class Query(gobject.GObject):
    
    __gsignals__ = {
        "started" : (gobject.SIGNAL_RUN_LAST,
                     gobject.TYPE_NONE,
                     tuple()),
        "finished" : (gobject.SIGNAL_RUN_LAST,
                      gobject.TYPE_NONE,
                      tuple())
    }
    
    def __init__(self, statement, cursor):
        self.__gobject_init__()
        self.statement = statement
        self.cursor = cursor
        self.description = None
        self.rowcount = -1
        self.rows = None
        self.messages = None
        self.failed = False
        self.executed = False
        self.execution_time = None
        self.coding_hint = "utf-8"
        self.errors = list()
        
    def execute(self, threaded=False):
        if threaded:
            Emit(self, "started")
        else:
            self.emit("started")
        start = time.time()
        try:
            self.cursor.execute(self.statement)
        except:
            self.failed = True
            self.errors.append(str(sys.exc_info()[1]))
        self.executed = True
        self.execution_time = time.time() - start
        self.messages = self.cursor.get_messages()
        if not self.failed:
            self.description = self.cursor.description
            self.rowcount = self.cursor.rowcount
            if self.description:
                self.rows = self.cursor.fetchall()
        if threaded:
            Emit(self, "finished")
        else:
            self.emit("finished")
            
class ReferenceProvider(gobject.GObject):
    name = None
    base_url = None
    
    def __init__(self):
        self.__gobject_init__()
        
    def get_context_help_url(self, term):
        return None


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""SQL-specific TextView

This module provides a SQL-specific TextView widget. Depending on
your system it is a gtksourceview2 or gtksourceview implementation.

Both implementations inherit ``SQLViewBase`` which connects to the
configuration instance to update some options when they are changed
through the preferences dialog.

Example usage
=============

    .. sourcecode:: python
    
        >>> from cf.ui.widgets.sqlview import SQLView
        >>> import gtk
        >>> w = gtk.Window()
        >>> sw = gtk.ScrolledWindow()
        >>> view = SQLView(app)
        >>> view.get_buffer().set_text("select * from foo")
        >>> sw.add(view)
        >>> w.add(sw)
        >>> w.show_all()
        
"""

import gconf
import gtk
import gobject
import pango
   
class SQLViewBase(object):
    """Base class for SQLView implementations"""
    
    def __init__(self, app):
        """
        This constructor takes one argument:
        
        :Parameter:
            app
                `CFApplication`_ instance
                
        .. _CFApplication: cf.app.CFApplication.html
        """
        self.app = app
        self.app.config.connect("changed", self.on_config_changed)
        self.update_textview_options()
        
    def on_config_changed(self, config, option, value):
        if option.startswith("editor."):
            gobject.idle_add(self.update_textview_options)
            
    def update_textview_options(self):
        """Updates the textview settings"""
        conf = self.app.config
        buffer = self.get_buffer()
        if conf.get("editor.wrap_text"):
            if conf.get("editor.wrap_split"):
                self.set_wrap_mode(gtk.WRAP_CHAR)
            else:
                self.set_wrap_mode(gtk.WRAP_WORD)
        else:
            self.set_wrap_mode(gtk.WRAP_NONE)
        self.set_show_line_numbers(conf.get("editor.display_line_numbers"))
        self.set_highlight_current_line(conf.get("editor.highlight_current_line"))
        self.set_insert_spaces_instead_of_tabs(conf.get("editor.insert_spaces"))
        self.set_auto_indent(conf.get("editor.auto_indent"))
        if conf.get("editor.default_font"):
            client = gconf.client_get_default()
            font = client.get_string("/desktop/gnome/interface/monospace_font_name")
        else:
            font = conf.get("editor.font")
        self.modify_font(pango.FontDescription(font))
        
try:
    import gtksourceview2
    USE_GTKSOURCEVIEW2 = True
    class SQLView(gtksourceview2.View, SQLViewBase):
        """SQLViewBase implementation"""
        def __init__(self, app):
            gtksourceview2.View.__init__(self)
            buffer = gtksourceview2.Buffer()
            self.set_buffer(buffer)
            lm = gtksourceview2.language_manager_get_default()
            buffer.set_language(lm.get_language("sql"))
            SQLViewBase.__init__(self, app)
            
        def update_textview_options(self):
            SQLViewBase.update_textview_options(self)
            conf = self.app.config
            buffer = self.get_buffer()
            # gtksourceview2 specific
            self.set_show_right_margin(conf.get("editor.right_margin"))
            self.set_right_margin_position(conf.get("editor.right_margin_position"))
            buffer.set_highlight_matching_brackets(conf.get("editor.bracket_matching"))
            self.set_tab_width(conf.get("editor.tabs_width"))
            sm = gtksourceview2.style_scheme_manager_get_default()
            scheme = sm.get_scheme(conf.get("editor.scheme"))
            buffer.set_style_scheme(scheme)
            
except ImportError:
    USE_GTKSOURCEVIEW2 = False
    import gtksourceview
    class SQLView(gtksourceview.SourceView, SQLViewBase):
        """SQLViewBase implementation"""
        def __init__(self, app):
            gtksourceview.SourceView.__init__(self)
            buffer = gtksourceview.SourceBuffer()
            self.set_buffer(buffer)
            buffer.set_highlight(True)
            lm = gtksourceview.SourceLanguagesManager()
            for lang in lm.get_available_languages():
                if lang.get_id() == "SQL":
                    buffer.set_language(lang)
                    break
            SQLViewBase.__init__(self, app)
                
        def update_textview_options(self):
            SQLViewBase.update_textview_options(self)
            conf = self.app.config
            buffer = self.get_buffer()
            # gtksourceview specific
            self.set_show_margin(conf.get("editor.right_margin"))
            #self.set_right_margin(conf.get("editor.right_margin_position"))
            buffer.set_check_brackets(conf.get("editor.bracket_matching"))
            self.set_tabs_width(conf.get("editor.tabs_width"))


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: app.py 176 2008-04-29 18:27:05Z albrecht.andi $

"""Application object

"""

import gobject
import gtk

import sys

from cf import release
from config import Config 
from datasources import DatasourceManager
from plugins.core import PluginManager
from ui.prefs import PreferencesDialog
from userdb import UserDB
from dbus_manager import CFService

import logging
log = logging.getLogger("APP")

class CFApplication(gobject.GObject):
    """Main application object
    
    An instance of this class is accessible in almost every class in
    CrunchyFrog through the ``app`` attribute. It provides access
    to application internals.
    
    The easiest way to learn more about this object is to activate the
    Python shell plugin. The shell has two local variables: ``app``
    is the CFApplication instance and ``instance`` is the `CFInstance`_
    object for the GUI the shell runs in.
        
    Attributes
    ==========
    
        :config: Configuration (see `Config`_ class)
        :userdb: User database (see `UserDB`_ class)
        :plugins: Plugin registry (see `PluginManager`_ class)
        :datasources: Datasource information (see `DatasourceManager`_ class)
            
    .. _Config: cf.config.Config.html
    .. _UserDB: cf.userdb.UserDB.html
    .. _PluginManager: cf.plugins.core.PluginManager.html
    .. _DatasourceManager: cf.datasources.DatasourceManager.html
    .. _CFInstance: cf.instance.CFInstance.html
    """
    
    def __init__(self, options):
        """
        The constructor of this class takes one argument:
        
        :Parameter:
            options
                Commandline options as returned by `optparse`_
                
        .. _optparse: http://docs.python.org/lib/module-optparse.html
        """
        self.__gobject_init__()
        self.set_data("instances", {})
        self.__icount = 0L
        self.options = options
        
    def init(self):
        """Initializes the application"""
        self.cb = CFAppCallbacks()
        self.__shutdown_tasks = []
        self.config = Config(self, self.options.config)
        self.userdb = UserDB(self)
        self.plugins = PluginManager(self)
        self.datasources = DatasourceManager(self)
        self.dbus_service = CFService(self)
        self.recent_manager = gtk.recent_manager_get_default()
        
    def new_instance(self, args=None):
        """Creates a new instances.
        
        ``args`` is an optional list of file names to open
        """
        from cf.instance import CFInstance
        instance = CFInstance(self)
        instances = self.get_data("instances")
        self.__icount += 1
        instances[self.__icount] = instance
        self.set_data("instances", instances)
        instance.set_data("instance-id", self.__icount)
        instance.widget.connect("destroy", self.on_instance_destroyed, self.__icount)
        instance.widget.show()
        self.cb.emit("instance-created", instance)
        return instance
    
    def on_instance_destroyed(self, instance, instance_id):
        instances = self.get_data("instances")
        if instances.has_key(instance_id):
            del instances[instance_id]
        self.set_data("instances", instances)
        if not instances:
            self.shutdown()
            
    def get_instance(self, instance_id):
        """Returns an instances identified by ID"""
        return self.get_data("instances").get(instance_id)
        
    def get_instances(self):
        """Returns a list of active instances
        
        :Returns: List of `CFInstance`_ instances
        
        .. _CFInstance: cf.instance.CFInstance.html
        """
        return self.get_data("instances").values()
        
    def preferences_show(self):
        """Displays the preferences dialog"""
        dlg = PreferencesDialog(self)
        dlg.run() # IGNORE:E1101 - Generic method
        dlg.destroy() # IGNORE:E1101 - Generic method
        
    def register_shutdown_task(self, func, msg, *args, **kwargs):
        """Registers a new shutdown task.
        
        Shutdown tasks are executed when the ``shutdown`` method is
        called.
        
        :Parameter:
            func
                A callable
            msg
                A human readable message explaingin what the task does
            args, kwargs
                arguments and keyword arguments for ``func`` (optional)
        """
        self.__shutdown_tasks.append((func, msg, args, kwargs))
        
    def shutdown(self):
        """Execute all shutdown tasks and quit application
        
        Shutdown tasks are called within a try-except-block. If an
        exception is raised it will be printed to stdout.
        """
        while self.__shutdown_tasks:
            func, msg, args, kwargs = self.__shutdown_tasks.pop()
            log.info(msg)
            try:
                func(*args, **kwargs)
            except:
                import traceback; traceback.print_exc()
                log.error("Task failed: %s" % str(sys.exc_info()[1]))
        gtk.main_quit()
        
class CFAppCallbacks(gobject.GObject):
    """Container for application specific signals.
    
    Signals
    =======
    
        instance-created
            ``def callback(application, instance, user_data1, ...)``
            
            Called when a new instance was created
    """
    
    __gsignals__ = {
        "instance-created" : (gobject.SIGNAL_RUN_LAST,
                              gobject.TYPE_NONE,
                              (gobject.TYPE_PYOBJECT,)),
    }


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""LDAP backend"""

import gtk
import gobject
import pango

import sys

from cf.backends import DBConnection, DBConnectError
from cf.backends.schema import SchemaProvider, Node
from cf.plugins.core import DBBackendPlugin
from cf.datasources import DatasourceInfo
from cf.ui import GladeWidget
from cf.ui.pdock import DockItem
from cf.ui.widgets import DataExportDialog

from gettext import gettext as _


import logging
log = logging.getLogger("LDAP")

class LDAPBackend(DBBackendPlugin):
    id = "crunchyfrog.backend.ldap"
    name = _(u"LDAP Plugin")
    description = _(u"Provides access to LDAP/ActiveDirectory servers")
        
    def __init__(self, app):
        DBBackendPlugin.__init__(self, app)
        log.info("Activating LDAP backend")
        self.schema = LDAPSchema()
        self.app.cb.connect("instance-created", self.on_instance_created)
        for item in self.app.get_instances():
            self.init_instance(item)
            
    def on_instance_created(self, app, instance):
        self.init_instance(instance)
        
    def on_search_dn(self, menuitem, object, instance):
        search = LDAPSearch(self.app, instance, object.get_data("connection"))
        search.set_search_dn(object.get_data("dn"))
        item = DockItem(instance.dock, "search_%s" % object.get_data("dn"),
                        search.widget, 
                        _(u"Search: %(dn)s") % {"dn" : object.get_data("dn")}, 
                       "gtk-find", None)
        instance.dock.add_item(item)
        
    def init_instance(self, instance):
        instance.browser.connect("object-menu-popup", self.on_object_menu_popup, instance)
        
    def on_object_menu_popup(self, browser, popup, object, instance):
        if isinstance(object, LDAPNode):
            item = gtk.ImageMenuItem("gtk-find")
            item.show()
            item.connect("activate", self.on_search_dn, object, instance)
            popup.append(item)
            sep = gtk.SeparatorMenuItem()
            sep.show()
            popup.append(sep)
            item = gtk.MenuItem(_(u"_Details"))
            item.show()
            item.connect("activate", self.on_object_details, object, instance)
            popup.append(item)
            
    def on_object_details(self, menuitem, object, instance):
        res = object.get_data("connection").conn.search_s(object.get_data("dn"), ldap.SCOPE_BASE)
        view = LDAPView(res[0][1])
        item = DockItem(instance.dock, object.get_data("dn"), view, object.get_data("dn"), 
                       "gtk-edit", None)
        instance.dock.add_item(item)

    @classmethod
    def _get_basedn(cls, entry):
        return entry.get_text().strip() or None
        
    def shutdown(self):
        log.info("Shutting down LDAP backend")
    
    @classmethod
    def get_datasource_options_widgets(cls, data_widgets, initial_data=None):
        lbl = gtk.Label(_(u"Base DN:"))
        lbl.set_alignment(0, 0.5)
        entry_basedn = gtk.Entry()
        if initial_data and initial_data.options.get("base_dn"):
            entry_basedn.set_text(initial_data.options.get("base_dn"))
        data_widgets["base_dn"] = (cls._get_basedn, entry_basedn)
        return data_widgets, ["host", "port", lbl, entry_basedn, "user", "password"]
    
    @classmethod
    def get_label(cls, datasource_info):
        return "%s (%s)" % (datasource_info.name, datasource_info.options.get("host"))
    
    def dbconnect(self, data):
        try:
            import ldap
        except ImportError:
            raise DBConnectError(_(u"Python module ldap is not installed."))
        try:
            conn_args = [data.get("host") or "localhost"]
            conn_args.append(data.get("port") or 389)
            real_conn = ldap.open(*conn_args)
        except:
            raise DBConnectError(str(sys.exc_info()[1]))
        if data.get("user"):
            try:
                real_conn.bind_s(data.get("user"), data.get("password"))
            except:
                raise DBConnectError(str(sys.exc_info()[1]))
        return LDAPConnection(self, self.app, real_conn, data)
    
    def test_connection(self, data):
        try:
            conn = self.dbconnect(data)
            conn.close()
        except DBConnectError, err:
            return str(err)
        return None
    
class LDAPConnection(DBConnection):
    
    def __init__(self, provider, app, conn, opts):
        DBConnection.__init__(self, provider, app)
        self.conn = conn
        self.opts = opts
        
    def _reconnect(self):
        data = self.opts
        conn_args = [data.get("host") or "localhost"]
        conn_args.append(data.get("port") or 389)
        real_conn = ldap.open(*conn_args)
        if data.get("user"):
            try:
                real_conn.bind_s(data.get("user"), data.get("password"))
            except:
                raise DBConnectError(str(sys.exc_info()[1]))
        self.conn = real_conn
        
    def search_s(self, *args, **kw):
        try:
            r = self.conn.search_s(*args, **kw)
        except ldap.SERVER_DOWN:
            self._reconnect()
            r = self.conn.search_s(*args, **kw)
        return r
            
        
class LDAPSearch(GladeWidget):
    
    def __init__(self, app, instance, connection):
        GladeWidget.__init__(self, app, "crunchyfrog", "ldap_search")
        self.connection = connection
        self.instance = instance
        
    def _setup_connections(self):
        self.xml.get_widget("ldap_search_results").connect("button-press-event", self.on_button_press_event)
        
    def on_button_press_event(self, treeview, event):
        if event.button == 3:
            x = int(event.x)
            y = int(event.y)
            time = event.time
            pthinfo = treeview.get_path_at_pos(x, y)
            if pthinfo is not None:
                path, col, cellx, celly = pthinfo
                treeview.grab_focus()
                treeview.set_cursor( path, col, 0)
                popup = gtk.Menu()
                model = treeview.get_model()
                iter = model.get_iter(path)
                item = gtk.MenuItem(_(u"Export"))
                item.connect("activate", self.on_export)
                item.show()
                popup.append(item)
                popup.show_all()
                popup.popup( None, None, None, event.button, time)
            return 1
        
    def on_do_search(self, *args):
        self.search()
        
    def on_export(self, *args):
        gobject.idle_add(self.export_data)
        
    def on_row_activated(self, treeview, path, column):
        model = treeview.get_model()
        iter = model.get_iter(path)
        dn = model.get_value(iter, 0)
        res = self.connection.conn.search_s(dn, ldap.SCOPE_BASE)
        view = LDAPView(res[0][1])
        item = DockItem(self.instance.dock, dn, view, dn, 
                       "gtk-edit", None)
        self.instance.dock.add_item(item)
        
    def export_data(self):
        data = []
        treeview = self.xml.get_widget("ldap_search_results")
        model = treeview.get_model()
        iter = model.get_iter_first()
        description = []
        first_row = True
        while iter:
            row = []
            for i in range(len(treeview.get_columns())):
                row.append(model.get_value(iter, i))
                if first_row:
                    col = treeview.get_column(i)
                    description.append((col.get_title(), str, None, None, None, None, None))
            data.append(row)
            first_row = False
            iter = model.iter_next(iter)
        selected = None
        statement = ""
        gtk.gdk.threads_enter()
        dlg = DataExportDialog(self.app, self.instance.widget,
                               data, selected, statement,
                               description)
        if dlg.run() == gtk.RESPONSE_OK:
            dlg.hide()
            dlg.export_data()
        dlg.destroy()
        gtk.gdk.threads_leave()
        
    def search(self):
        search_dn = self.xml.get_widget("ldap_search_dn").get_text()
        if self.xml.get_widget("ldap_search_scope_onelevel").get_active():
            scope = ldap.SCOPE_ONELEVEL
        else:
            scope = ldap.SCOPE_SUBTREE
        filterstr = self.xml.get_widget("ldap_search_filter").get_text()
        attrlist = self.xml.get_widget("ldap_search_attributes").get_text().split(",") 
        r = self.connection.search_s(search_dn, scope, filterstr, attrlist)
        treeview = self.xml.get_widget("ldap_search_results")
        while treeview.get_columns():
            treeview.remove_column(treeview.get_column(0))
        model_args = [str]
        renderer = gtk.CellRendererText()
        renderer.set_property("ellipsize", pango.ELLIPSIZE_END)
        col = gtk.TreeViewColumn("dn", renderer, text=0)
        col.set_sort_column_id(0)
        col.set_resizable(True)
        col.set_min_width(100)
        col.set_reorderable(True)
        treeview.append_column(col)
        for i in range(len(attrlist)):
            attr = attrlist[i]
            model_args.append(str)
            renderer = gtk.CellRendererText()
            renderer.set_property("ellipsize", pango.ELLIPSIZE_END)
            col = gtk.TreeViewColumn(attr, renderer, text=i+1)
            col.set_sort_column_id(i+1)
            col.set_resizable(True)
            col.set_reorderable(True)
            col.set_min_width(50)
            treeview.append_column(col)
        model = gtk.ListStore(*model_args)
        treeview.set_model(model)
        for item in r:
            row = [item[0]]
            attrs = item[1]
            for i in range(len(attrlist)):
                att = attrlist[i].lower()
                att_set = False
                for key, value in attrs.items():
                    if key.lower() == att and not att_set:
                        row.append("\n".join(value))
                        att_set = True
                if not att_set:
                    row.append("N/A") 
            model.append(row)
        
        
    def set_search_dn(self, dn):
        self.xml.get_widget("ldap_search_dn").set_text(dn)
        
class LDAPView(gtk.ScrolledWindow):
    
    def __init__(self, data):
        gtk.ScrolledWindow.__init__(self)
        self.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        self.model = gtk.ListStore(str, str, str)
        self.list = gtk.TreeView(self.model)
        col = gtk.TreeViewColumn(_(u"Attribute"), gtk.CellRendererText(), text=0)
        col.set_sort_column_id(2)
        self.list.append_column(col)
        col = gtk.TreeViewColumn(_(u"Value"), gtk.CellRendererText(), text=1)
        col.set_sort_column_id(1)
        self.list.append_column(col)
        for key, value in data.items():
            for i in range(len(value)):
                if i == 0:
                    self.model.append([key, value[i], key])
                else:
                    self.model.append(["", value[i], key])
        self.add(self.list)
        self.list.set_rules_hint(True)
        
class LDAPNode(Node):
    icon = "gtk-open"
        
class LDAPSchema(SchemaProvider):
    
    def fetch_children(self, connection, parent):
        if isinstance(parent, DatasourceInfo):
            dn = connection.opts.get("base_dn")
        else:
            dn = parent.get_data("dn")
        ret = []
        for item in  [x[0] for x in connection.search_s(dn, ldap.SCOPE_ONELEVEL, attrsonly=1)]:
            node = LDAPNode(item.split(",", 1)[0], dn=item, connection=connection)
            if len(connection.search_s(item, ldap.SCOPE_ONELEVEL, attrsonly=1)) == 0:
                node.has_children = False
                node.icon = "gtk-justify-fill"
            ret.append(node)
        return ret


try:
    import ldap
except ImportError:
    LDAPBackend.INIT_ERROR = _(u"Python module ldap required.")


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: utils.py 48 2008-03-03 16:12:29Z albrecht.andi $

"""Nifty little helpers"""

import gtk
import gobject

# found here: http://www.mail-archive.com/pygtk@daa.com.au/msg12556.html
def Emit(obj, *args):

    def Do_Emit(obj, sig, *args):
        gtk.gdk.threads_enter()
        obj.emit(sig, *args)
        gtk.gdk.threads_leave()
        return False
    gobject.idle_add(Do_Emit, obj, *args, **{"priority" : gobject.PRIORITY_HIGH})
#END: Emit


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""Mixin classes for plugins

Plugins can use this classes to provide a certain behaviour
or to have special attributes. For example, if a plugin needs to
add an item to the menubar of an instance, the plugin must connect
to the ``instance-created`` signal and have to fetch the menubar from
the instances widget tree. The `MenubarMixin`_ provides a 
``menubar_load(menubar, instance)`` method which is automatically called
by the `plugin manager`_. So you don't need to pay attention on instance
creation and can focus on the menubar itself when your plugin implements
these methods.

.. _MenubarMixin: cf.plugins.mixins.MenubarMixin.html
.. _plugin manager: cf.plugins.core.PluginManager.html
"""

class InstanceMixin(object):
    """Instance relation
    
    This mixin can be used by plugins that require a relation to a 
    running instance.
    """
    
    def init_instance(self, instance):
        """Called on instance creation.
        
        :Parameter:
            instance
                the created instance
        """
        pass
    
class MenubarMixin(InstanceMixin):
    """Modify the applications main menu
    
    ``menubar_load`` is called on activation for each instance and
    if a new instance is created.
    
    ``menubar_unload`` is called when the plugin is deactivated for 
    each instance.
    """
    
    def menubar_load(self, menubar, instance):
        """Add items to the main menu
        
        :Parameter:
            menubar
                Main menu (``gtk.MenuBar``)
            instance
                An instance
        """
        pass
    
    def menubar_unload(self, menubar, instance):
        """Remove items to the main menu
        
        :Parameter:
            menubar
                Main menu (``gtk.MenuBar``)
            instance
                An instance
        """
        pass
    
class EditorMixin(InstanceMixin):
    """Editor relation
    
    This mixin can be used to keep track of the current editor.
    """
    
    def set_editor(self, editor, instance):
        """Called when an editor gets activated
        
        :Parameter:
            editor
                Active editor or ``None``
            instance
                An instance
        """
        pass
    
class UserDBMixin:
    """User database access
    
    For more information about how use the user database see `UserDB`_.
    
    Sets ``userdb`` property.
    
    .. _`UserDB`: cf.userdb.UserDB.html
    """
    
    def userdb_set(self, userdb):
        """Associates the userdb with the plugin
        
        .. Warning:: This method shouldn't be subclassed by plugins.
        """
        self.set_data("userdb", userdb)
    
    def userdb_get(self):
        """Returns the associated userdb"""
        return self.get_data("userdb")
    
    userdb = property(fget=userdb_get)
    
    def userdb_init(self):
        """Called when a plugin is activated
        
        This is the right place to create tables, check table versions
        and doing table upgrades.
        """
        pass


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: browser.py 129 2008-03-26 09:06:40Z albrecht.andi $

"""Object browser"""

import gtk
import gobject

from kiwi.ui import dialogs

from cf.datasources import DatasourceInfo
from cf.backends import DBConnectError
from cf.backends.schema import Collection
from cf.ui import pdock
from cf.ui.editor import SQLView

from gettext import gettext as _

class DummyNode(object):
    pass

class Browser(gtk.ScrolledWindow):
    
    __gsignals__ = {
        "object-menu-popup" : (gobject.SIGNAL_RUN_LAST,
                               gobject.TYPE_NONE,
                               (gobject.TYPE_PYOBJECT,
                                gobject.TYPE_PYOBJECT)),
    }
    
    def __init__(self, app, instance):
        self.app = app
        self.instance = instance
        self._setup_widget()
        self._setup_connections()
        self.reset_tree()
        
    def _setup_widget(self):
        gtk.ScrolledWindow.__init__(self)
        self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.set_size_request(250, -1)
        # Model
        self.model = gtk.TreeStore(gobject.TYPE_PYOBJECT, # 0 Object
                                   str, # 1 label
                                   gtk.gdk.Pixbuf, # 2 icon
                                   gobject.TYPE_PYOBJECT, # 3 double-click callback
                                   gobject.TYPE_PYOBJECT, # 4 popup menu callback
                                   str, # 5 tooltip
                                   )
        self.model.set_sort_column_id(1, gtk.SORT_ASCENDING)
        # TreeView
        self.object_tree = gtk.TreeView(self.model)
        self.add(self.object_tree)
        self.object_tree.set_tooltip_column(5)
        col = gtk.TreeViewColumn()
        renderer = gtk.CellRendererPixbuf()
        col.pack_start(renderer, expand=False)
        col.add_attribute(renderer, 'pixbuf', 2)
        renderer = gtk.CellRendererText()
        col.pack_start(renderer, expand=True)
        col.add_attribute(renderer, 'text', 1)
        self.object_tree.append_column(col)
        self.object_tree.set_headers_visible(False)
        self.object_tree.enable_model_drag_source(gtk.gdk.BUTTON1_MASK, (("text/plain", gtk.TARGET_SAME_APP, 80),), gtk.gdk.ACTION_DEFAULT | gtk.gdk.ACTION_COPY)
        def drag_data_get(treeview, context, selection, info, timestamp):
            treeselection = treeview.get_selection()
            model, iter = treeselection.get_selected()
            text = id(model.get_value(iter, 0))
            selection.set('text/plain', 8, str(text))
            return
        self.object_tree.connect("drag_data_get", drag_data_get)
        self.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        
    def _setup_connections(self):
        self.app.datasources.connect("datasource-added", self.on_datasource_added)
        self.app.datasources.connect("datasource-deleted", self.on_datasource_deleted)
        self.app.datasources.connect("datasource-modified", self.on_datasource_modified)
        self.object_tree.connect("button-press-event", self.on_button_press_event)
        self.object_tree.connect("row-expanded", self.on_row_expanded)
        sel = self.object_tree.get_selection()
        sel.connect("changed", self.on_object_tree_selection_changed)
        
    def _create_dsinfo_menu(self, model, iter, popup):
        def connect(item, datasource_info):
            try:
                datasource_info.dbconnect()
                self.on_object_tree_selection_changed(self.object_tree.get_selection())
            except DBConnectError, err:
                dialogs.error(_(u"Connection failed"), str(err))
        def disconnect(item, datasource_info):
            datasource_info.dbdisconnect()
            iter = self.get_iter_for_datasource(datasource_info)
            model = self.object_tree.get_model()
            self.object_tree.collapse_row(model.get_path(iter))
        datasource_info = model.get_value(iter, 0)
        if datasource_info.get_connections():
            lbl = _(u"Disconnect")
            cb = disconnect
        else:
            lbl = _(u"Connect")
            cb = connect
        item = gtk.MenuItem(lbl)
        item.connect("activate", cb, datasource_info)
        popup.append(item)
        return popup
        
    def on_button_press_event(self, treeview, event):
        if event.type == gtk.gdk._2BUTTON_PRESS \
        and event.button == 1:
            x = int(event.x)
            y = int(event.y)
            time = event.time
            pthinfo = treeview.get_path_at_pos(x, y)
            if pthinfo is not None:
                path, col, cellx, celly = pthinfo
                treeview.grab_focus()
                treeview.set_cursor( path, col, 0)
                model = treeview.get_model()
                iter = model.get_iter(path)
                obj = model.get_value(iter, 0)
                if isinstance(obj, DatasourceInfo) \
                and not obj.internal_connection:
                    self.instance.statusbar.set_message(_(u"Connecting..."))
                    try:
                        conn = obj.dbconnect()
                        editor = self.instance.new_editor()
                        editor.set_connection(conn)
                        gobject.idle_add(editor.textview.grab_focus)
                        self.on_object_tree_selection_changed(self.object_tree.get_selection())
                    except DBConnectError, err:
                        dialogs.error(_(u"Connection failed"), str(err))
                        self.instance.statusbar.set_message("")
                elif isinstance(obj, DatasourceInfo):
                    editor = self.instance.new_editor()
                    editor.set_connection(obj.internal_connection)
                    gobject.idle_add(editor.textview.grab_focus)
        elif event.button == 3:
            x = int(event.x)
            y = int(event.y)
            time = event.time
            pthinfo = treeview.get_path_at_pos(x, y)
            if pthinfo is not None:
                path, col, cellx, celly = pthinfo
                treeview.grab_focus()
                treeview.set_cursor( path, col, 0)
                popup = gtk.Menu()
                model = treeview.get_model()
                iter = model.get_iter(path)
                obj = model.get_value(iter, 0)
                if model.iter_children(iter):
                    item = gtk.ImageMenuItem("gtk-refresh")
                    item.connect("activate", self.on_refresh_node, model, iter)
                    popup.append(item)
                if obj.has_details:
                    item = gtk.MenuItem(_(u"Details"))
                    item.connect("activate", self.on_show_details, obj, model, iter)
                    popup.append(item)
                if popup.get_children():
                    popup.append(gtk.SeparatorMenuItem())
                cb = model.get_value(iter, 4)
                if cb:
                    try:
                        popup = cb(model, iter, popup)
                    except:
                        import traceback; traceback.print_exc()
                self.emit("object-menu-popup", popup, model.get_value(iter, 0))
                if popup.get_children():
                    popup.show_all()
                    popup.popup( None, None, None, event.button, time)
            return 1
        
    def on_datasource_added(self, manager, datasource_info):
        iter = self.model.append(None)
        self.set_datasource_info(iter, datasource_info)
    
    def on_datasource_deleted(self, manager, datasource_info):
        iter = self.get_iter_for_datasource(datasource_info)
        if not iter:
            return
        self.model.remove(iter)
    
    def on_datasource_modified(self, manager, datasource_info):
        iter = self.get_iter_for_datasource(datasource_info)
        if not iter:
            return
        self.set_datasource_info(iter, datasource_info)
        
    def on_object_tree_selection_changed(self, selection):
        model, iter = selection.get_selected()
        if not iter: return
        obj = model.get_value(iter, 0)
        if isinstance(obj, DatasourceInfo):
            if obj.internal_connection:
                server_info = obj.internal_connection.get_server_info()
                if server_info:
                    self.instance.statusbar.set_message(server_info)
                else:
                    self.instance.statusbar.set_message("")
            else:
                self.instance.statusbar.set_message("")
        else:
            comm = model.get_value(iter, 5) or ""
            self.instance.statusbar.set_message(comm)
            
    def on_refresh_node(self, menuitem, model, iter):
        citer = model.iter_children(iter)
        while citer:
            model.remove(citer)
            citer = model.iter_children(iter)
        citer = model.append(iter)
        model.set_value(citer, 0, DummyNode())
        self.object_tree.emit("row-expanded", iter, model.get_path(iter))
        
    def on_row_expanded(self, treeview, iter, path):
        model = treeview.get_model()
        obj = model.get_value(iter, 0)
        citer = model.iter_children(iter)
        if citer:
            cobj = model.get_value(citer, 0)
            if isinstance(cobj, DummyNode):
                datasource_info = self.find_datasource_info(model, citer)
                model.remove(citer)
                if datasource_info.backend.schema and datasource_info.internal_connection:
                    for child in datasource_info.backend.schema.fetch_children(datasource_info.internal_connection, obj) or []:
                        citer = model.append(iter)
                        if child.icon:
                            icon = gtk.icon_theme_get_default().load_icon(child.icon, gtk.ICON_SIZE_MENU, gtk.ICON_LOOKUP_FORCE_SVG)
                        else:
                            icon = None
                        model.set(citer, 
                                  0, child,
                                  1, child.name,
                                  2, icon,
                                  5, child.description)
                        if child.has_children:
                            cciter = model.append(citer)
                            model.set(cciter, 0, DummyNode())
                    citer = model.iter_children(iter)
                    if citer:
                        treeview.expand_row(model.get_path(iter), False)
                        
    def on_show_details(self, menuitem, object, model, iter):
        datasource_info = self.find_datasource_info(model, iter)
        if datasource_info.backend.schema:
            data = datasource_info.backend.schema.get_details(datasource_info.internal_connection, object)
            if not data:
                return
            nb = gtk.Notebook()
            for key, value in data.items():
                if isinstance(value, basestring):
                    widget = SQLView(self.app)
                    sw = gtk.ScrolledWindow()
                    sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
                    sw.add(widget)
                    widget.set_editable(False)
                    widget.get_buffer().set_text(value)
                    nb.append_page(sw, gtk.Label(key))
            nb.show_all()
            item = pdock.DockItem(self.instance.dock, "details", nb, _(u"Details: %(objname)s") % {"objname": object.name},
                              "gtk-edit", None)
            self.instance.dock.add_item(item)
            item.show_all()
                    
    def find_datasource_info(self, model, iter):
        while iter:
            obj = model.get_value(iter, 0)
            if isinstance(obj, DatasourceInfo):
                return obj
            iter = model.iter_parent(iter)
    
    def get_iter_for_datasource(self, datasource_info):
        iter = self.model.get_iter_first()
        while iter:
            if self.model.get_value(iter, 0) == datasource_info:
                return iter
            iter = self.model.iter_next(iter)
        return None
    
    def get_object_by_id(self, id_, iter=None, model=None):
        if not iter:
            model = self.object_tree.get_model()
            iter = model.get_iter_first()
        while iter:
            value = model.get_value(iter, 0)
            if id(value) == id_:
                return value
            citer = model.iter_children(iter)
            while citer:
                ret = self.get_object_by_id(id_, citer, model)
                if ret:
                    return ret
                citer = model.iter_next(citer)
            iter = model.iter_next(iter)
    
    def reset_tree(self):
        self.model.clear()
        for datasource_info in self.app.datasources.get_all():
            iter = self.model.append(None)
            self.set_datasource_info(iter, datasource_info)
    
    def set_datasource_info(self, iter, datasource_info):
        if datasource_info.get_connections():
            ico = gtk.icon_theme_get_default().load_icon("stock_connect", gtk.ICON_SIZE_MENU, gtk.ICON_LOOKUP_FORCE_SVG)
        else:
            ico = gtk.icon_theme_get_default().load_icon("stock_disconnect", gtk.ICON_SIZE_MENU, gtk.ICON_LOOKUP_FORCE_SVG)
        self.model.set(iter,
                  0, datasource_info,
                  1, datasource_info.get_label(),
                  2, ico,
                  3, None,
                  4, self._create_dsinfo_menu,
                  5, datasource_info.description)
        citer = self.model.iter_children(iter)
        if datasource_info.get_connections() \
        and not citer:
            citer = self.model.append(iter)
            self.model.set(citer, 0, DummyNode())
        elif not datasource_info.get_connections() \
        and citer:
            self.model.remove(citer)


"""apachelogs.py: code for reading and parsing Apache log files.
Based heavily on apachelogs.py, authored by Kevin Scott (kevin.scott@gmail.com)

Sample use -- counting the number of 40xs seen:

  import apachelogs

  alf = ApacheLogFile('access.log')
  40x_count = 0
  for log_line in alf:
    if log_line.http_response_code.startswith('40'):
      40x_count += 1
  alf.close()
  print "Saw %d 40x responses" % 40x_count
"""


import re
import time
import urlparse
import cgi
import datetime
import os
import mmap
import fileinput

class ApacheLogLine:
  """A Python class whose attributes are the fields of Apache log line.

  Designed specifically with combined format access logs in mind.  For
  example, the log line

  127.0.0.1 - frank [10/Oct/2000:13:55:36 -0700] "GET /apache_pb.gif HTTP/1.0" 200 2326 "http://www.example.com/start.html" "Mozilla/4.08 [en] (Win98; I ;Nav)"

  would have the following field values as an ApacheLogLine:

  ip = '127.0.0.1'
  ident = '-'
  http_user = 'frank'
  time = '10/Oct/2000:13:55:36 -0700'
  request_line = 'GET /apache_pb.gif HTTP/1.0'
  http_response_code = '200'
  http_response_size = 2326
  referrer = 'http://www.example.com/start.html'
  user_agent = 'Mozilla/4.08 [en] (Win98; I ;Nav)'
  http_method = 'GET'
  url = '/apache_pb.gif'
  http_vers = 'HTTP/1.0'
  """
  def __init__(self, *rgroups ):
    self.ip, self.ident, \
    self.http_user, self.time, \
    self.request_line, self.http_response_code, \
    self.http_response_size, self.referrer, self.user_agent = rgroups
    self.http_method, self.url, self.http_vers = self.request_line.split()
    
  def __str__(self):
    """Return a simple string representation of an ApacheLogLine."""
    return ' '.join([self.ip, self.ident, self.time, self.request_line,
        self.http_response_code, self.http_response_size, self.referrer,
        self.user_agent])


_lineRegex = re.compile(r'(\d+\.\d+\.\d+\.\d+) ([^ ]*) ([^ ]*) \[([^\]]*)\] "([^"]*)" (\d+) ([^ ]*) "([^"]*)" "([^"]*)"')

class ApacheLogFile:
  """An abstraction for reading and parsing Apache log files."""
  def __init__(self, *filename):
    """Instantiating an ApacheLogFile opens a log file.  
    Client is responsible for closing the opened log file by calling close()"""
    self.f = fileinput.input(filename)

  def close(self):
    """Closes the Apache log file."""
    self.f.close()

  def __iter__(self):
    """Returns a log line object for each iteration. """
    match = _lineRegex.match
    for line in self.f:
      m = match(line)
      if m:
        try:
          log_line = ApacheLogLine(*m.groups())
          yield log_line
        except:
          print "NON_COMPLIANT_FORMAT: ", line
          
        



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""Django Backend
"""

import gtk

from cf.plugins.core import DBBackendPlugin, PLUGIN_TYPE_BACKEND
from cf.backends import DBConnection, DBCursor, DBConnectError
from cf.datasources import DatasourceInfo
from cf.plugins.mixins import MenubarMixin, EditorMixin
from cf.ui.widgets import CustomImageMenuItem
from cf.backends.schema import *

import os
import sys
import md5
from gettext import gettext as _

import logging
log = logging.getLogger("DJANGO")


class DjangoBackendPlugin(DBBackendPlugin):
    id = "crunchyfrog.backend.django"
    plugin_type = PLUGIN_TYPE_BACKEND
    name = "Django Backend"
    description = "Provides access to models in a Django project"
    author = "Andi Albrecht"
    license = "GPL"
    homepage = "http://cf.andialbrecht.de"
    version = "0.1"
    
    def __init__(self, *args, **kw):
        DBBackendPlugin.__init__(self, *args, **kw)
        self.schema = DjangoSchema()
        
    @classmethod
    def _get_filename(cls, chooser):
        return chooser.get_filename()
        
    def shutdown(self):
        log.info("Shutting down Django backend")
    
    @classmethod
    def get_datasource_options_widgets(cls, data_widgets, initial_data=None):
        lbl = gtk.Label(_(u"Settings file:"))
        lbl.set_alignment(0, 0.5)
        file_chooser = gtk.FileChooserButton(_(u"Select Django settings file"))
        filter = gtk.FileFilter()
        filter.set_name("Django settings")
        filter.add_pattern("settings.py")
        file_chooser.add_filter(filter)
        file_chooser.set_filter(filter)
        if initial_data:
            file_chooser.select_filename(initial_data.options.get("settings_file"))
        data_widgets["settings_file"] = (cls._get_filename, file_chooser)
        return data_widgets, [lbl, file_chooser]
    
    @classmethod
    def get_label(cls, datasource_info):
        return "%s (django://%s)" % (datasource_info.name, datasource_info.options.get("settings_file"))
    
    def dbconnect(self, data):
        try:
            import django
        except ImportError:
            raise DBConnectError(_(u"Python module django is not installed."))
        return DjangoConnection(self, self.app, data)
    
    def test_connection(self, data):
        try:
            conn = self.dbconnect(data)
            conn.close()
        except DBConnectError, err:
            return str(err)
        return None
    
class DjangoCursor(DBCursor):
    
    def __init__(self, connection):
        DBCursor.__init__(self, connection)
        self._cur = self.connection._conn.cursor()
        
    def _get_description(self):
        return self._cur.cursor.description
    description = property(fget=_get_description)
    
    def _get_rowcount(self):
        return self._cur.cursor.rowcount
    rowcount = property(fget=_get_rowcount)
        
    def execute(self, statement):
        try:
            self._cur.execute(statement)
            exc = None
        except:
            exc = sys.exc_info()
        self.connection.update_transaction_status()
        if exc:
            raise exc[1]
        
    def fetchall(self):
        return self._cur.cursor.fetchall()
    
    def close(self):
        self._cur.cursor.close()
    
class DjangoConnection(DBConnection):
    
    cursor_class = DjangoCursor
    
    def __init__(self, provider, app, data):
        DBConnection.__init__(self, provider, app)
        self.data = data
        self.settings_file = data["settings_file"]
        self.settings_path = os.path.dirname(self.settings_file)
        self.app_name = os.path.basename(self.settings_path)
        self.settings_env = "CF_DJANGO_%s" % md5.new(self.settings_file).hexdigest()
        sys.path.insert(0, os.path.abspath(os.path.join(self.settings_path, "../")))
        os.environ[self.settings_env] = "%s.settings" % (self.app_name)
        from django import conf
        conf.ENVIRONMENT_VARIABLE = self.settings_env
        from django.conf import settings
        self.settings = settings
        from django.db import connection
        self._conn = connection
        
    def close(self):
        self._conn.close()
        DBConnection.close(self)
        
    def cursor(self):
        return self.cursor_class(self)
    
    def commit(self):
        self._conn._commit()
        self.update_transaction_status()
        
    def rollback(self):
        self._conn._rollback()
        self.update_transaction_status()
        
    def explain(self, statement):
        try:
            sql = "EXPLAIN %s" % statement
            cur = self._conn.cursor()
            cur.execute(sql)
            data = cur.fetchall()
            cur.close()
        except:
            gtk.gdk.threads_enter()
            dialogs.error(_(u"An error occured"), str(sys.exc_info()[1]))
            gtk.gdk.threads_leave()
            return []
        return data
        
class DjApp(Node):
    icon = "gtk-open"
    
class DjModel(Table):
    pass
    
class DjModelField(Column):
    pass
        
class DjangoSchema(SchemaProvider):
    
    def __init__(self):
        SchemaProvider.__init__(self)
        
    def fetch_children(self, connection, parent):
        from django.db import models
        if isinstance(parent, DatasourceInfo):
            return [DjApp(app.__name__.split('.')[-2],
                          _(u"Module: %s") % app.__name__,
                          app=app) for app in models.get_apps()]
        elif isinstance(parent, DjApp):
            return [DjModel(model._meta.object_name,
                            _(u"Table: %s") % model._meta.db_table,
                            model=model) for model in models.get_models(parent.get_data("app"))]
        elif isinstance(parent, DjModel):
            return [DjModelField(field.name,
                                 _(u"Column: %s (%s)") % (field.get_attname(),
                                                          field.db_type()),
                                 field=field)
                    for field in parent.get_data("model")._meta.fields]


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: __init__.py 149 2008-03-28 07:05:02Z albrecht.andi $

"""
CrunchyFrog Package

A good starting point to read this API documentation is the `CFApplication`_
class.


Other Resources
===============
    Development pages
        http://cf.andialbrecht.de
    Additional documentation
        http://code.google.com/p/crunchyfrog/wiki/Documentation?tm=6
    Discussions
        http://groups.google.com/group/crunchyfrog

.. _GNOME: http://www.gnome.org
.. _Wiki: http://code.google.com/p/crunchyfrog/wiki/Documentation?tm=6
.. _CFApplication: cf.app.CFApplication.html
"""

__doc_all__ = ["main", "backends"]

import pygtk; pygtk.require("2.0")

import bonobo
import gnome
import gnomevfs
import gobject
import gtk
import gtk.glade
import gettext
import traceback
import os
import errno

gobject.threads_init()

from cf import release

from optparse import OptionParser
from os.path import abspath, dirname, join, isfile, isdir, expanduser
from os import makedirs
import sys

#if sys.version_info[:2] >= (2, 5):
#    LOG_FORMAT_APP = '[%(levelname)s] %(asctime)s %(message)s\n\t%(pathname)s:%(lineno)s in %(funcName)s\n\tPID: %(process)s, Thread: %(threadName)s [%(thread)d], Logger: %(name)s\n'
#else:
#    LOG_FORMAT_APP = '[%(levelname)s] %(asctime)s %(message)s\n\t%(pathname)s:%(lineno)s\n\tPID: %(process)s, Thread: %(threadName)s [%(thread)d], Logger: %(name)s\n'
LOG_FORMAT_APP = '%(levelname)s\t%(name)s\t%(created)f\t%(message)s'

import logging
logging.basicConfig(format=LOG_FORMAT_APP)

if isfile(abspath(join(dirname(__file__), "../setup.py"))):
    DATA_DIR = abspath(join(dirname(__file__), "../data"))
    LOCALE_DIR = abspath(join(dirname(__file__), "../po"))
else:
    from dist import DATA_DIR, LOCALE_DIR
PLUGIN_DIR = join(DATA_DIR, "plugins")
USER_CONFIG_DIR = abspath(expanduser("~/.config/crunchyfrog"))
USER_CONF = join(USER_CONFIG_DIR, "config")
if not isdir(USER_CONFIG_DIR):
    makedirs(USER_CONFIG_DIR)
USER_DIR = abspath(expanduser("~/.crunchyfrog"))
if not isdir(USER_DIR):
    makedirs(USER_DIR)
USER_PLUGIN_DIR = join(USER_DIR, "plugins/")
if not isdir(USER_PLUGIN_DIR):
    makedirs(USER_PLUGIN_DIR)
USER_PLUGIN_URI = gnomevfs.get_uri_from_local_path(USER_PLUGIN_DIR)
USER_PLUGIN_REPO = join(USER_DIR, "repo.xml")
USER_PLUGIN_REPO_URI = gnomevfs.get_uri_from_local_path(USER_PLUGIN_REPO)
PID_FILE = join(USER_DIR, "crunchyfrog.pid")
    
gettext.bindtextdomain("crunchyfrog", LOCALE_DIR)
gettext.textdomain("crunchyfrog")
gtk.glade.bindtextdomain("crunchyfrog", LOCALE_DIR)
gtk.glade.textdomain("crunchyfrog")

from cf.app import CFApplication
from cf import dbus_manager

def _parse_commandline():
    """Parses command line arguments and handles all arguments
    which exit immediately (e.g. --version)"""
    usage = "usage: %prog [options] FILE1, FILE2, ..."
    parser = OptionParser(usage)
    parser.add_option("-d", "--debug",
                      action="store_true", dest="debug",
                      default=False,
                      help="run in debug mode")
    parser.add_option("--version",
                      action="store_true", dest="show_version",
                      default=False,
                      help="show program's version number and exit")
    parser.add_option("-c", "--config",
                      dest="config", default=USER_CONF,
                      help="configuration file")
    options, args = parser.parse_args()
    if options.show_version:
        print "%s - %s" % (release.name, release.description)
        print "Version %s" % release.version
        print
        sys.exit()
    return options, args

# grabbed from listen / gajm
def is_alive(pid_file):
    try:
        pf = open(pid_file)
    except:
        # probably file not found
        return False

    try:
        pid = int(pf.read().strip())
        pf.close()
    except:
        traceback.print_exc()
        # PID file exists, but something happened trying to read PID
        # Could be 0.10 style empty PID file, so assume Gajim is running
        return False

    try:
        if not os.path.exists('/proc'):
            print "missing /proc"
            return True # no /proc, assume Listen is running
        try:
            f = open('/proc/%d/cmdline'% pid) 
        except IOError, e:
            if e.errno == errno.ENOENT:
                return False # file/pid does not exist
            raise 

        n = f.read().lower()
        f.close()
        if n.find('cf/__init__.py') < 0:
            return False
        return True # Running Listen found at pid
    except:
        traceback.print_exc()

    # If we are here, pidfile exists, but some unexpected error occured.
    # Assume Listen is running.
    return True


def main():
    global FIRST_RUN
    options, args = _parse_commandline()
    options.first_run = not isfile(options.config)
    if options.debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.WARNING
    logger = logging.getLogger()
    logger.setLevel(log_level)
    if not is_alive(PID_FILE):
        if isfile(abspath(join(dirname(__file__), "../setup.py"))):
            props = {'app-datadir': abspath(join(dirname(__file__), '../data'))}
        else:
            props = dict()
        gnome.init(release.name.lower(), release.version,
                   properties=props)       
        app = CFApplication(options)
        app.init()
        instance = app.new_instance(args)
        f = open(PID_FILE, "w")
        f.write(str(os.getpid()))
        f.close()
        logger.debug("PID: %d [%s]" % (os.getpid(), PID_FILE))
        for fname in args:
            instance.new_editor(fname)
        gtk.main()
        os.remove(PID_FILE)
    else:
        client = dbus_manager.get_client()
        if args:
            from cf.instance import InstanceSelector
            sel = InstanceSelector(client)
            if sel.run() == 1:
                instance_id = sel.get_instance_id()
                if not instance_id:
                    instance_id = client.new_instance()
                for fname in args:
                    client.open_uri(instance_id, fname)
            sel.destroy()
        else:
            client.new_instance()
            
        
if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: queries.py 179 2008-05-07 15:18:24Z albrecht.andi $

"""Notebook that holds SQL editors"""

import gtk
import gobject
import pango

import re
import os

from gettext import gettext as _

class QueriesNotebook(gtk.Notebook):
    
    def __init__(self, app, instance):
        self.app = app
        self.instance = instance
        gtk.Notebook.__init__(self)
        self.set_scrollable(True)
        self.popup_disable()
        self.set_property("enable-popup", False)
        self.connect("switch-page", self.on_switch_page)
        self.connect("page-removed", self.on_page_removed)
        self.connect("page-added", self.on_page_added)
        
    def on_page_added(self, notebook, child, page_num):
        gobject.idle_add(self.set_current_page, page_num)
        
    def on_page_removed(self, notebook, child, page_num):
        gobject.idle_add(self.instance.set_editor_active, child, False)
        
    def on_switch_page(self, notebook, page, page_num):
        editor = self.get_nth_page(page_num).get_data("glade-widget")
        gobject.idle_add(self.instance.set_editor_active, editor, True)
            
    def attach(self, editor):
        if editor.get_parent():
            editor.reparent(self)
            self.set_tab_label(editor.widget, TabLabel(editor))
        else: 
            self.append_page(editor.widget, TabLabel(editor))
        self.set_tab_reorderable(editor.widget, True)
        self.popup_disable()
        self.set_property("enable-popup", False)
        
class TabLabel(gtk.HBox):
    
    def __init__(self, editor):
        gtk.HBox.__init__(self)
        self.label = gtk.Label(_(u"Query"))
        self.label.set_ellipsize(pango.ELLIPSIZE_END)
        self.label.set_width_chars(15)
        self.label.set_single_line_mode(True)
        self.label.set_alignment(0, 0.5)
        self.editor = editor
        self.editor.connect("connection-changed", self.on_editor_connection_changed)
        buffer = self.editor.textview.get_buffer()
        buffer.connect("changed", self.on_buffer_changed)
        self.update_label(buffer)
        self.pack_start(self.label, True, True)
        btn_close = gtk.Button()
        btn_close.connect("clicked", self.on_button_close_clicked)
        btn_close.connect("button-press-event", self.on_button_press_event)
        self.add_icon_to_button(btn_close)
        btn_close.set_relief(gtk.RELIEF_NONE)
        self.pack_start(btn_close, False, False)
        self.update_tooltip()
        self.show_all()
        
    def add_icon_to_button(self,button):
        iconBox = gtk.HBox(False, 0)        
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_CLOSE,gtk.ICON_SIZE_MENU)
        gtk.Button.set_relief(button,gtk.RELIEF_NONE)
        #gtk.Button.set_focus_on_click(button,False)
        settings = gtk.Widget.get_settings (button)
        (w,h) = gtk.icon_size_lookup_for_settings(settings,gtk.ICON_SIZE_MENU)
        gtk.Widget.set_size_request (button, w+4, h+4);
        image.show()
        iconBox.pack_start(image, True, False, 0)
        button.add(iconBox)
        iconBox.show()
        return 
    
    def on_button_close_clicked(self, button):
        self.editor.close()
        
    def on_buffer_changed(self, buffer):
        gobject.idle_add(self.update_label, buffer)
        
    def on_button_press_event(self, box, event):
        if event.button == 3:
            time = event.time
            popup = gtk.Menu()
            item = gtk.MenuItem(_(u"Show in separate window"))
            item.connect("activate", self.on_show_in_separate_window)
            popup.append(item)
            sep = gtk.SeparatorMenuItem()
            sep.show()
            popup.append(sep)
            item = gtk.ImageMenuItem("gtk-close")
            item.connect("activate", self.on_close_editor)
            item.show()
            popup.append(item)
            popup.show_all()
            popup.popup( None, None, None, event.button, time)
            
    def on_close_editor(self, *args):
        self.editor.close()
        
    def on_editor_connection_changed(self, editor, connection):
        self.update_tooltip()
            
    def on_show_in_separate_window(self, item):
        gobject.idle_add(self.editor.show_in_separate_window)
        
    def update_label(self, buffer):
        if self.editor.get_filename():
            fname = self.editor.get_filename()
            txt = os.path.split(fname)[-1]
            self.label.set_tooltip_text(fname)
            if self.editor.file_contents_changed():
                txt = "*"+txt
        else:
            self.label.set_tooltip_text("")
            txt = buffer.get_text(*buffer.get_bounds())
            txt = re.sub("\s+", " ", txt)
        txt = txt.strip()
        if not txt:
            txt = _(u"Query")
        self.label.set_text(txt)
        self.update_tooltip()
        
    def update_tooltip(self):
        markup = "<b>Connection:</b> "
        if self.editor.connection:
            markup += self.editor.connection.get_label()
        else:
            markup += "["+_(u"Not connected")+"]"
        if self.editor.get_filename():
            markup += "\n<b>File:</b> "+self.editor.get_filename()
        self.label.set_tooltip_markup(markup)


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

import gtk
import gobject

import os

from gettext import gettext as _

from cf.ui import GladeWidget

MODE_SINGLE = 1
MODE_MULTIPLE = 2

class ConfirmSaveDialog(GladeWidget):
    
    def __init__(self, instance, changed_editors):
        GladeWidget.__init__(self, instance.app, "crunchyfrog",
                             "saveconfirm_dialog")
        self.changed_editors = changed_editors
        if len(changed_editors) == 1:
            self.set_dialog_mode(MODE_SINGLE)
        else:
            self.set_dialog_mode(MODE_MULTIPLE)
        self.fill_filelist()
        
    def _setup_widget(self):
        model = gtk.ListStore(gobject.TYPE_PYOBJECT, str, bool, str)
        self.filelist = self.xml.get_widget("filelist")
        self.filelist.set_model(model)
        self.filelist.set_tooltip_column(3)
        col = gtk.TreeViewColumn()
        renderer = gtk.CellRendererToggle()
        renderer.connect("toggled", self.on_entry_toggled)
        col.pack_start(renderer, expand=False)
        col.add_attribute(renderer, "active", 2)
        renderer = gtk.CellRendererText()
        col.pack_start(renderer, expand=True)
        col.add_attribute(renderer, 'text', 1)
        self.filelist.append_column(col)
        
    def on_entry_toggled(self, renderer, path):
        model = self.filelist.get_model()
        iter = model.get_iter(path)
        model.set_value(iter, 2, not model.get_value(iter, 2))
        
    def fill_filelist(self):
        model = self.filelist.get_model()
        for i in range(len(self.changed_editors)):
            editor = self.changed_editors[i]
            if editor.get_filename():
                lbl = os.path.basename(editor.get_filename())
                tip = None
            else:
                lbl = _(u"Unsaved Query %(num)s") % {"num" : i+1}
                content = list()
                lines = editor.get_text().splitlines()
                for line in lines:
                    if len(line) > 35:
                        content.append(line[:35]+"...")
                    else:
                        content.append(line)
                    if len(content) == 7:
                        if len(lines) > 7:
                            content.append("...")
                        break
                tip = "\n".join(content)
            model.append([editor, lbl, True, tip])
        
    def set_dialog_mode(self, mode):
        if mode == MODE_SINGLE:
            self.xml.get_widget("lbl_filelist").hide()
            self.xml.get_widget("sw_filelist").hide()
            self.xml.get_widget("filelist").hide()
            editor = self.changed_editors[0]
            if editor.get_filename():
                filename = os.path.basename(editor.get_filename())
            else:
                filename = _(u"Unsaved Query %(num)s") % {"num" : 1} 
            msg_header = _(u'Save the changes to document "%(filename)s" before closing?') % {"filename":filename}
        else:
            self.xml.get_widget("lbl_filelist").show()
            self.xml.get_widget("sw_filelist").show()
            msg_header = _(u"There are %(num)d documents with unsaved changes. Save changes before closing?")
            msg_header = msg_header % {"num" : len(self.changed_editors)}
        lbl = self.xml.get_widget("lbl_header").get_label()
        self.xml.get_widget("lbl_header").set_markup(lbl % msg_header)
        
    def save_files(self):
        model = self.filelist.get_model()
        iter = model.get_iter_first()
        while iter:
            if model.get_value(iter, 2):
                if not model.get_value(iter, 0).save_file(default_name=model.get_value(iter, 1)):
                    return False
            iter = model.iter_next(iter)
        return True


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""SQL formatter plugin

The printsql module is based on the work of Peter Bengtsson.
Visit his blog for details: http://www.peterbe.com/Pretty-print-SQL-script
"""

import gtk

from cf.plugins.core import GenericPlugin, PLUGIN_TYPE_EDITOR
from cf.plugins.mixins import MenubarMixin, EditorMixin
from cf.ui.widgets import CustomImageMenuItem

from gettext import gettext as _

from printsql import printsql

class SQLFormatterPlugin(GenericPlugin, MenubarMixin, EditorMixin):
    id = "cf.editor.sqlformat"
    plugin_type = PLUGIN_TYPE_EDITOR
    name = "SQL Formatter"
    description = "Formats the statement in current editor"
    icon = "stock_autoformat"
    author = "Andi Albrecht"
    license = "GPL"
    homepage = "http://cf.andialbrecht.de"
    version = "0.1"
    
    def __init__(self, *args, **kw):
        GenericPlugin.__init__(self, *args, **kw)
        self._mn_items = dict()
    
    def menubar_load(self, menubar, instance):
        edit_menu = instance.xml.get_widget("mn_edit").get_submenu()
        sep = gtk.SeparatorMenuItem()
        sep.show()
        edit_menu.append(sep)
        item = CustomImageMenuItem("stock_autoformat", _(u"Format SQL"))
        item.set_sensitive(False)
        item.connect("activate", self.on_format_sql, instance)
        item.show()
        edit_menu.append(item)
        self._mn_items[instance] = [sep, item]
        
    def menubar_unload(self, menubar, instance):
        for item in self._mn_items[instance]:
            item.get_parent().remove(item)
            item.destroy()
        del self._mn_items[instance]
        
    def set_editor(self, editor, instance):
        if not self._mn_items.has_key(instance):
            return
        self._mn_items[instance][1].set_sensitive(bool(editor))
        
    def on_format_sql(self, menuitem, instance):
        editor = instance.get_editor()
        if not editor:
            return
        sql = editor.get_text()
        editor.set_text(printsql(sql))


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: editor.py 176 2008-04-29 18:27:05Z albrecht.andi $

"""SQL editor and results view"""

import gtk
import gobject
import gconf
import pango
import gnome
import gnomevfs

import re
import thread
import time
import os
import string

from gettext import gettext as _
from kiwi.ui import dialogs

from cf.backends import Query
from cf.ui import GladeWidget
from cf.ui.toolbar import CFToolbar
from cf.ui.widgets import DataExportDialog
from cf.ui.widgets.grid import Grid
from cf.ui.widgets.sqlview import SQLView
from cf.ui.confirmsave import ConfirmSaveDialog

    

class Editor(GladeWidget):
    
    __gsignals__ = {
        "connection-changed" : (gobject.SIGNAL_RUN_LAST,
                                gobject.TYPE_NONE,
                                (gobject.TYPE_PYOBJECT,))
    }
    
    def __init__(self, app, instance):
        self.app = app
        self.instance = instance
        self.connection = None
        self.__conn_close_tag = None
        self._query_timer = None
        self._filename = None
        self._filecontent_read = ""
        GladeWidget.__init__(self, app, "crunchyfrog", "box_editor")
        self.set_data("win", None)
        self.show_all()
        
    def _setup_widget(self):
        self._setup_textview()
        self._setup_resultsgrid()
        self.lbl_status = self.xml.get_widget("editor_label_status")
        
    def _setup_textview(self):
        self.textview = SQLView(self.app)
        sw = self.xml.get_widget("sw_editor_textview")
        sw.add(self.textview)
        
    def _setup_resultsgrid(self):
        self.results = ResultsView(self.app, self.instance, self.xml)
        
    def _setup_connections(self):
        self.textview.connect("populate-popup", self.on_populate_popup)
        
    def on_close(self, *args):
        self.close()
            
    def on_connection_closed(self, connection):
        if connection == self.connection and self.__conn_close_tag:
            connection.disconnect(self.__conn_close_tag)
        self.set_connection(None)
        
    def on_explain(self, *args):
        gobject.idle_add(self.explain)
        
    def on_populate_popup(self, textview, popup):
        sep = gtk.SeparatorMenuItem()
        sep.show()
        popup.append(sep)
        if self.connection and self.connection.provider.reference:
            refviewer = self.instance.get_data("refviewer")
            buffer = self.textview.get_buffer()
            bounds = buffer.get_selection_bounds()
            if bounds:
                url = self.connection.provider.reference.get_context_help_url(buffer.get_text(*bounds))
            else:
                url = None
            if url and refviewer:
                item = gtk.ImageMenuItem("gtk-help")
                item.connect("activate", self.on_show_context_help, refviewer, url)
                item.show()
                popup.append(item)
        if self.get_data("win"):
            lbl = _(u"Show in main window")
            cb = self.on_show_in_main_window
        else:
            lbl = _(u"Show in separate window")
            cb = self.on_show_in_separate_window
        item = gtk.MenuItem(lbl)
        item.connect("activate", cb)
        item.show()
        popup.append(item)
        item = gtk.ImageMenuItem("gtk-close")
        item.show()
        item.connect("activate", self.on_close)
        popup.append(item)
        
    def on_query_started(self, query):
        start = time.time()
        # Note: The higher the time out value, the longer the query takes.
        #    50 is a reasonable value anyway.
        #    The start time is for the UI only. The real execution time
        #    is calculated in the Query class.
        self._query_timer = gobject.timeout_add(50, self.update_exectime, start, query)
        
    def on_query_finished(self, query, tag_notice):
        self.results.set_query(query)
        if query.failed:
            gobject.idle_add(self.lbl_status.set_text, _(u"Query failed (%.3f seconds)") % query.execution_time)
        elif query.description:
            gobject.idle_add(self.lbl_status.set_text, _(u"Query finished (%.3f seconds, %s rows)") % (query.execution_time, query.rowcount))
        else:
            gobject.idle_add(self.lbl_status.set_text, _(u"Query finished (%.3f seconds, %s affected rows)") % (query.execution_time, query.rowcount))
        self.connection.disconnect(tag_notice)
        gobject.idle_add(self.textview.grab_focus)
        
    def on_show_context_help(self, menuitem, refviewer, url):
        refviewer.load_url(url)
        
    def on_show_in_main_window(self, *args):
        gobject.idle_add(self.show_in_main_window)
        
    def on_show_in_separate_window(self, *args):
        gobject.idle_add(self.show_in_separate_window)
        
    def close(self):
        dlg = ConfirmSaveDialog(self, [self])
        resp = dlg.run()
        if resp == 1:
            ret = dlg.save_files()
        elif resp == 2:
            ret = True
        else:
            ret = False
        dlg.destroy()
        if ret:
            if self.get_data("win"):
                self.get_data("win").destroy()
            else:
                self.destroy()
            self.instance._editors.remove(self)
            return True
        
    def commit(self):
        if not self.connection: return
        cur = self.connection.cursor()
        cur.execute("commit")
        cur.close()
        
    def rollback(self):
        if not self.connection: return
        cur = self.connection.cursor()
        cur.execute("rollback")
        cur.close()
        
    def begin_transaction(self):
        if not self.connection: return
        cur = self.connection.cursor()
        cur.execute("begin")
        cur.close()
        
    def execute_query(self):
        self.lbl_status.set_text(_(u"Executing query..."))
        def exec_threaded(statement):
            cur = self.connection.cursor() 
            query = Query(statement, cur)
            query.coding_hint = self.connection.coding_hint
            gtk.gdk.threads_enter()
            query.connect("started", self.on_query_started)
            query.connect("finished", self.on_query_finished, tag_notice)
            gtk.gdk.threads_leave()
            query.execute(True)
        buffer = self.textview.get_buffer()
        bounds = buffer.get_selection_bounds()
        self.results.reset()
        if not bounds:
            bounds = buffer.get_bounds()
        statement = buffer.get_text(*bounds)
        if self.app.config.get("editor.replace_variables"):
            tpl = string.Template(statement)
            tpl_search = tpl.pattern.search(tpl.template)
            if tpl_search and tpl_search.groupdict().get("named"):
                dlg = StatementVariablesDialog(tpl)
                gtk.gdk.threads_enter()
                if dlg.run() == gtk.RESPONSE_OK:
                    statement = dlg.get_statement()
                else:
                    statement = None
                dlg.destroy()
                gtk.gdk.threads_leave()
                if not statement:
                    self.lbl_status.set_text(_(u"Query cancelled"))
                    return
        def foo(connection, msg):
            self.results.add_message(msg)
        tag_notice = self.connection.connect("notice", foo)
        if self.connection.threadsafety >= 2:
            thread.start_new_thread(exec_threaded, (statement,))
        else:
            cur = self.connection.cursor()
            query = Query(statement, cur)
            query.coding_hint = self.connection.coding_hint
            query.connect("finished", self.on_query_finished, tag_notice)
            query.execute()
            
    def explain(self):
        buffer = self.textview.get_buffer()
        bounds = buffer.get_selection_bounds()
        if not bounds:
            bounds = buffer.get_bounds()
        statement = buffer.get_text(*bounds)
        data = []
        if self.connection:
            data = self.connection.explain(statement)
        self.results.set_explain(data)
        
    def set_connection(self, conn):
        if self.connection and self.__conn_close_tag:
            self.connection.disconnect(self.__conn_close_tag)
            self.__conn_close_tag = None
        self.connection = conn
        if conn:
            self.__conn_close_tag = self.connection.connect("closed", self.on_connection_closed)
        else:
            self._conn_close_tag = None
        self.emit("connection-changed", conn)
        
    def set_filename(self, filename):
        self._filename = filename
        if filename:
            f = open(self._filename)
            a = f.read()
            f.close()
        else:
            a = ""
        self._filecontent_read = a
        self.textview.get_buffer().set_text(a)
        self.app.recent_manager.add_item(gnomevfs.get_uri_from_local_path(filename))
        
    def get_filename(self):
        return self._filename
    
    def file_contents_changed(self):
        if self._filename:
            buffer = self.textview.get_buffer()
            return buffer.get_text(*buffer.get_bounds()) != self._filecontent_read
        return False
    
    def contents_changed(self):
        if self._filename:
            return self.file_contents_changed()
        elif len(self.get_text()) == 0:
            return False
        return True
    
    def file_confirm_save(self):
        dlg = dialogs.yesno(_(u"Save file %(name)s before closing the editor?") % {"name":os.path.basename(self._filename)})
        if dlg == gtk.RESPONSE_YES:
            return self.save_file_as()
        return True
    
    def save_file(self, parent=None, default_name=None):
        if not self._filename:
            return self.save_file_as(parent=parent, default_name=default_name)
        buffer = self.get_buffer()
        a = buffer.get_text(*buffer.get_bounds())
        f = open(self._filename, "w")
        f.write(a)
        f.close()
        self._filecontent_read = a
        gobject.idle_add(buffer.emit, "changed")
        return True
        
    def save_file_as(self, parent=None, default_name=None):
        if not parent:
            parent = self.instance.widget
        dlg = gtk.FileChooserDialog(_(u"Save file"),
                            parent,
                            gtk.FILE_CHOOSER_ACTION_SAVE,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        if self._filename:
            dlg.set_filename(self._filename)
        else:
            dlg.set_current_folder(self.app.config.get("editor.recent_folder", ""))
            if default_name:
                dlg.set_current_name(default_name)
        filter = gtk.FileFilter()
        filter.set_name(_(u"All files (*)"))
        filter.add_pattern("*")
        dlg.add_filter(filter)
        filter = gtk.FileFilter()
        filter.set_name(_(u"SQL files (*.sql)"))
        filter.add_pattern("*.sql")
        dlg.add_filter(filter)
        dlg.set_filter(filter)
        if dlg.run() == gtk.RESPONSE_OK:
            self._filename = dlg.get_filename()
            gobject.idle_add(self.save_file)
            self.app.config.set("editor.recent_folder", dlg.get_current_folder())
            ret = True
        else:
            ret = False
        dlg.destroy()
        return ret
    
    def get_buffer(self):
        return self.textview.get_buffer()
    
    def get_text(self):
        buffer = self.get_buffer()
        return buffer.get_text(*buffer.get_bounds())
    
    def set_text(self, txt):
        buffer = self.get_buffer().set_text(txt)
        
    def show_in_separate_window(self):
        win = EditorWindow(self.app, self.instance)
        win.attach(self)
        win.show_all()
        self.set_data("win", win)
        
    def show_in_main_window(self):
        self.instance.queries.attach(self)
        win = self.get_data("win")
        if win:
            win.destroy()
        self.set_data("win", None)
        
    def update_exectime(self, start, query):
        self.lbl_status.set_text("Query running... (%.3f seconds)" % (time.time()-start))
        if query.executed:
            gobject.source_remove(self._query_timer)
            self._query_timer = None
            if query.failed:
                gobject.idle_add(self.lbl_status.set_text, _(u"Query failed (%.3f seconds)") % query.execution_time)
            elif query.description:
                gobject.idle_add(self.lbl_status.set_text, _(u"Query finished (%.3f seconds, %s rows)") % (query.execution_time, query.rowcount))
            else:
                gobject.idle_add(self.lbl_status.set_text, _(u"Query finished (%.3f seconds, %s affected rows)") % (query.execution_time, query.rowcount))
            return False
        else:
            return True

        
class EditorWindow(GladeWidget):
    
    def __init__(self, app, instance):
        GladeWidget.__init__(self, app, "crunchyfrog", "editorwindow")
        self.instance = instance
        self.editor = None
        self.toolbar = CFToolbar(self.app, "crunchyfrog", cb_provider=self)
        box = self.xml.get_widget("mainbox_editor")
        box.pack_start(self.toolbar.widget, False, False)
        box.reorder_child(self.toolbar.widget, 1)
        quit = self.toolbar.xml.get_widget("tb_quit")
        self.toolbar.widget.remove(quit)
        item = gtk.ToolButton()
        item.set_label(_(u"Show in main window"))
        item.set_stock_id("gtk-dnd-multiple")
        item.connect("clicked", self.on_show_in_main_window)
        item.show()
        self.toolbar.insert(item, -1)
        item = gtk.ToolButton()
        item.set_stock_id("gtk-close")
        item.connect("clicked", self.on_close)
        item.show()
        self.toolbar.insert(item, -1)
        self.toolbar.set_icon_size(gtk.ICON_SIZE_MENU)
        self.restore_window_state()
        
    def on_buffer_changed(self, buffer):
        gobject.idle_add(self.update_title)
        
    def on_close(self, *args):
        self.close()
        
    def on_execute_query(self, *args):
        gobject.idle_add(self.editor.execute_query)
        
    def on_begin_transaction(self, *args):
        gobject.idle_add(self.editor.begin_transaction)
        
    def on_commit(self, *args):
        gobject.idle_add(self.editor.commit)
        
    def on_rollback(self, *args):
        gobject.idle_add(self.editor.rollback)
        
    def on_query_new(self, *args):
        self.instance.on_query_new(self, *args)
        
    def on_open_file(self, *args):
        dlg = gtk.FileChooserDialog(_(u"Select file"),
                            self.widget,
                            gtk.FILE_CHOOSER_ACTION_OPEN,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dlg.set_current_folder(self.app.config.get("editor.recent_folder", ""))
        filter = gtk.FileFilter()
        filter.set_name(_(u"All files (*)"))
        filter.add_pattern("*")
        dlg.add_filter(filter)
        filter = gtk.FileFilter()
        filter.set_name(_(u"SQL files (*.sql)"))
        filter.add_pattern("*.sql")
        dlg.add_filter(filter)
        dlg.set_filter(filter)
        if dlg.run() == gtk.RESPONSE_OK:
            gobject.idle_add(self.editor.set_filename, dlg.get_filename())
            self.app.config.set("editor.recent_folder", dlg.get_current_folder())
        dlg.destroy()
        
    def on_save_file(self, *args):
        self.editor.save_file(parent=self.widget)
        
    def on_save_file_as(self, *args):
        self.editor.save_file_as(parent=self.widget)
        
    def on_copy(self, *args):
        self.editor.textview.get_buffer().copy_clipboard(gtk.clipboard_get())
        
    def on_paste(self, *args): 
        self.editor.textview.get_buffer().paste_clipboard(gtk.clipboard_get(), None, True)
        
    def on_cut(self, *args):
        self.editor.textview.get_buffer().cut_clipboard(gtk.clipboard_get(), True)
        
    def on_delete(self, *args):
        self.editor.textview.get_buffer().delete_selection(True, True)
        
    def on_show_in_main_window(self, *args):
        self.editor.show_in_main_window()
        
    def on_configure_event(self, win, event):
        config = self.app.config
        if not config.get("querywin.maximized"):
            config.set("querywin.width", event.width)
            config.set("querywin.height", event.height)
            
    def on_window_state_event(self, win, event):
        config = self.app.config
        bit = gtk.gdk.WINDOW_STATE_MAXIMIZED.value_names[0] in event.new_window_state.value_names
        config.set("querywin.maximized", bit)
        
    def attach(self, editor):
        self.editor = editor
        box = self.xml.get_widget("mainbox_editor")
        ebox = self.xml.get_widget("box_editor")
        ebox.get_parent().remove(ebox)
        if self.editor.get_parent():
            self.editor.reparent(box)
            expand, fill, padding, pack_type = box.query_child_packing(self.editor.widget)
            box.set_child_packing(self.editor.widget, True, True, padding, gtk.PACK_START)
        else:
            box.pack_start(self.editor.widget, True, True)
        box.reorder_child(self.editor.widget, 2)
        buffer = self.editor.textview.get_buffer()
        buffer.connect("changed", self.on_buffer_changed)
        self.toolbar.set_editor(editor)
        self.update_title()
        
    def close(self):
        self.editor.close()
        
    def restore_window_state(self):
        if self.app.config.get("querywin.width", -1) != -1:
            self.widget.resize(self.app.config.get("querywin.width"),
                               self.app.config.get("querywin.height"))
        if self.app.config.get("querywin.maximized", False):
            self.widget.maximize()
        
    def update_title(self):
        buffer = self.editor.textview.get_buffer()
        txt = buffer.get_text(*buffer.get_bounds())
        txt = re.sub("\s+", " ", txt)
        txt = txt.strip()
        if len(txt) > 28:
            txt = txt[:25]+"..."
        if txt:
            txt = _(u"Query") + ": "+txt
        else:
            txt = _(u"Query")
        self.set_title(txt)
        

class ResultsView(GladeWidget):
    
    def __init__(self, app, instance, xml):
        self.instance = instance
        GladeWidget.__init__(self, app, xml, "editor_results")
        
    def _setup_widget(self):
        self.grid = ResultList(self.app, self.instance, self.xml)
        self.messages = self.xml.get_widget("editor_results_messages")
        buffer = self.messages.get_buffer()
        buffer.create_tag("error", foreground="#a40000", weight=pango.WEIGHT_BOLD)
        self.explain_model = gtk.ListStore(str)
        treeview = self.xml.get_widget("editor_explain")
        treeview.set_model(self.explain_model)
        col = gtk.TreeViewColumn("", gtk.CellRendererText(), text=0)
        treeview.append_column(col)
        
    def _setup_connections(self):
        self.grid.grid.connect("selection-changed", self.on_grid_selection_changed)
        
    def on_copy_data(self, *args):
        gobject.idle_add(self.copy_data)
        
    def on_export_data(self, *args):
        gobject.idle_add(self.export_data)
        
    def on_grid_selection_changed(self, grid, selected_cells):
        self.xml.get_widget("editor_copy_data").set_sensitive(bool(selected_cells))
        
    def copy_data(self):
        rows = dict()
        for cell in self.grid.grid.get_selected_cells():
            value = self.grid.grid.get_cell_data(cell, repr=True)
            if rows.has_key(cell[0]):
                rows[cell[0]] += "%s\t" % value
            else:
                rows[cell[0]] = "%s\t" % value
        rownums = rows.keys()
        rownums.sort()
        txt = ""
        for rownum in rownums:
            txt += "%s\n" % rows.get(rownum)
        display = gtk.gdk.display_manager_get().get_default_display()
        clipboard = gtk.Clipboard(display, "CLIPBOARD")
        clipboard.set_text(txt)
        self.instance.statusbar.set_message(_(u"Data copied to clipboard"))
        
    def export_data(self):
        data = self.grid.grid.get_grid_data()
        description = self.grid.grid.description
        selected = self.grid.grid.get_selected_rows()
        statement = self.grid.query.statement
        gtk.gdk.threads_enter()
        dlg = DataExportDialog(self.app, self.instance.widget, 
                               data, selected, statement, description)
        if dlg.run() == gtk.RESPONSE_OK:
            dlg.hide()
            dlg.export_data()
        dlg.destroy()
        gtk.gdk.threads_leave()
        
    def reset(self):
        # Explain
        self.explain_model.clear()
        # Messages
        buffer = self.messages.get_buffer()
        buffer.set_text("")
        self.set_current_page(2)
        
    def set_explain(self, data):
        self.explain_model.clear()
        for item in data:
            iter = self.explain_model.append()
            self.explain_model.set(iter, 0, item)
    
    def set_query(self, query):
        self.grid.set_query(query)
        buffer = self.messages.get_buffer()
        for err in query.errors:
            iter = buffer.get_end_iter()
            buffer.insert_with_tags_by_name(iter, err.strip()+"\n", "error")
        for msg in query.messages:
            buffer.insert_at_cursor(msg.strip()+"\n")
        if query.errors:
            curr_page = 2
        elif query.description:
            curr_page = 0
        else:
            curr_page = 2
        self.xml.get_widget("editor_export_data").set_sensitive(bool(query.rows))
        gobject.idle_add(self.set_current_page, curr_page)
        
    def add_message(self, msg):
        buffer = self.messages.get_buffer()
        buffer.insert_at_cursor(msg.strip()+"\n")



class ResultList(GladeWidget):
    """Result list with toolbar"""
    
    def __init__(self, app, instance, xml):
        GladeWidget.__init__(self, app, xml, "editor_results_data")
        self.instance = instance
        self.query = None
        
    def _setup_widget(self):
        self.grid = Grid()
        self.xml.get_widget("sw_grid").add(self.grid)
        
    def set_query(self, query):
        self.query = query
        self.grid.reset()
        if self.query.description:
            self.grid.set_result(self.query.rows, self.query.description,
                                 self.query.coding_hint)

            

class StatementVariablesDialog(gtk.Dialog):
    
    def __init__(self, template):
        gtk.Dialog.__init__(self, _(u"Variables"),
                            None,
                            gtk.DIALOG_MODAL|gtk.DIALOG_DESTROY_WITH_PARENT,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_OK, gtk.RESPONSE_OK))
        self.template = template
        self._widgets = dict()
        self._setup_widget()
        
    def _setup_widget(self):
        vars = []
        found = []
        for match in self.template.pattern.finditer(self.template.template):
            name = match.groupdict().get("named")
            if not name or name.lower() in found:
                continue
            vars.append(name)
            found.append(name.lower()) 
        table = gtk.Table(len(vars), 2)
        table.set_row_spacings(5)
        table.set_col_spacings(7)
        for i in range(len(vars)):
            lbl = gtk.Label(vars[i])
            lbl.set_alignment(0, 0.5)
            table.attach(lbl, 0, 1, i, i+1, gtk.FILL, gtk.FILL)
            entry = gtk.Entry()
            table.attach(entry, 1, 2, i, i+1, gtk.EXPAND|gtk.FILL, gtk.FILL)
            self._widgets[vars[i]] = entry
        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        sw.add_with_viewport(table)
        sw.set_border_width(10)
        self.vbox.pack_start(sw, True, True)
        self.vbox.show_all()
        
    def on_value_edited(self, renderer, path, value):
        model = self.treeview.get_model()
        iter = model.get_iter(path)
        model.set_value(iter, 1, value)
        
    def get_statement(self):
        data = dict()
        for var, widget in self._widgets.items():
            data[var] = widget.get_text()
        return self.template.safe_substitute(data)


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: __init__.py 66 2008-03-07 06:36:15Z albrecht.andi $

"""Plugins

This package provides base classes for writing plugins. The `core`_ module
contains very basic classes, like `GenericPlugin`_ - the base class for all 
plugins. `mixins`_ has some mixin classes for common tasks.

Related documents
=================
 * `Plugin HowTo`_
 
.. _core: cf.plugins.core.html
.. _mixins: cf.plugins.mixins.html
.. _GenericPlugin: cf.plugins.core.GenericPlugin.html
.. _Plugin HowTo: http://code.google.com/p/crunchyfrog/wiki/PluginDevelopment
"""



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""Python shell plugin"""

import gtk
import gobject

from cf.backends.schema import Table
from cf.plugins.core import GenericPlugin
from cf.plugins.mixins import InstanceMixin
from cf.ui.pdock import DockItem

from gettext import gettext as _

from ipython_view import *

class CFShell(GenericPlugin, InstanceMixin):
    id = "crunchyfrog.plugin.cfshell"
    name = _(u"Shell")
    description = _(u"Interactive shell (mainly for debugging)")
    icon = "gnome-terminal"
    author = "Andi Albrecht"
    license = "GPL"
    homepage = "http://cf.andialbrecht.de"
    version = "0.1"
    
    def __init__(self, app):
        GenericPlugin.__init__(self, app)
        self._shells = dict()
        self._instances = dict()
        for instance in app.get_instances():
            self.init_instance(instance)
            
    def on_import_table_to_shell(self, menuitem, object, view):
        view.import_table(object)
               
    def on_object_menu_popup(self, browser, popup, object, view):
        if isinstance(object, Table):
            item = gtk.MenuItem(_(u"Import to shell"))
            item.connect("activate", self.on_import_table_to_shell, object, view)
            item.show()
            popup.append(item)
        
    def on_toggle_shell(self, menuitem, instance):
        if menuitem.get_active():
            view = CFShellView(self.app, instance)
            self._shells[instance] = view
            item = DockItem(instance.dock, "cfshell", view, _(u"Shell"), 
                              "gnome-terminal", gtk.POS_BOTTOM)
            instance.dock.add_item(item)
            view.connect("destroy", self.on_view_destroyed, menuitem, instance)
            tag = instance.browser.connect("object-menu-popup", self.on_object_menu_popup, view)
            view.set_data("object-menu-tag", tag)
        else:
            instance.browser.disconnect(self._shells[instance].get_data("object-menu-tag"))
            self._shells[instance].destroy()
            del self._shells[instance]
    
    def on_view_destroyed(self, view, menuitem, instance):
        menuitem.set_active(False)
        
    def init_instance(self, instance):
        if instance in self._instances.keys():
            return
        mn_view = instance.xml.get_widget("mn_view")
        item = gtk.CheckMenuItem(_(u"Shell"))
        item.connect("activate", self.on_toggle_shell, instance)
        item.show()
        if self.app.config.get("cfshell.visible"):
            item.set_active(True)
        mn_view.append(item)
        self._instances[instance] = item
        
    def shutdown(self):
        if self._shells.keys():
            self.app.config.set("cfshell.visible", True)
        else:
            self.app.config.set("cfshell.visible", False)
        while self._instances:
            instance, mn_item = self._instances.popitem()
            mn_item.destroy()
        while self._shells:
            instance, shell = self._shells.popitem()
            shell.destroy()
        
class CFShellView(gtk.ScrolledWindow):
    
    def __init__(self, app, instance):
        gtk.ScrolledWindow.__init__(self)
        self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        self.app = app
        self.instance = instance
        self.iview = IPythonView()
        self.iview.connect("drag_data_received", self.on_drag_data_received)
        self.iview.drag_dest_set(gtk.DEST_DEFAULT_MOTION|gtk.DEST_DEFAULT_HIGHLIGHT|gtk.DEST_DEFAULT_DROP,
                                 [("text/plain", 0, 80)], gtk.gdk.ACTION_DEFAULT|gtk.gdk.ACTION_COPY)
        self.iview.updateNamespace({"app" : self.app, 
                                    "instance" : self.instance})
        self.add(self.iview)
        self.set_size_request(-1, 100)
        self.show_all()
        
    def on_drag_data_received(self, widget, context, x, y, selection, target_type, timestamp):
        object = self.instance.browser.get_object_by_id(int(selection.data))
        if isinstance(object, Table):
            self.import_table(object)
            context.drop_finish(True, timestamp)
        else:
            context.drop_finish(False, timestamp)
        widget.stop_emission("drag-data-received")
        return False
        
    def import_table(self, table):
        # TODO: Create some nice ORM and some usable models.
        #       Maybe we can fix this, when DDL operations are implemented...
        model_name = table.name.lower()+"_model"
        self.iview.updateNamespace({model_name : table})
        gobject.idle_add(self.instance.statusbar.set_message, _(u"Table %(tablename)r imported as %(varname)r to shell") % {"tablename" : table.name, "varname" : table.name.lower()})


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""A grid view"""

# NOTE:
#     This module should have no cf dependencies!

import gtk
import gobject
import pango
import gnomevfs

import sys
import os

from gettext import gettext as _

GRID_LABEL_MAX_LENGTH = 100

class Grid(gtk.TreeView):
    """Data grid
    
    The view component of the grid.
    
    Signals
    =======
    
        selection-changed
            ``def callback(grid, selected_cells, user_param1, ...)``
            
            Emitted when selected cells have changed
            
    
    Selection
    =========
    
    This grid allows three different selections which exclude each
    other:
    
        * one or more columns (select all means all columns are selected)
        * a single cell
        * one or more rows
        
    Columns can be selected by clicking on the column header. A click on the
    header of the first column selects or de-selects all data. Rows can be
    selected by clicking on the first column of a row. Individual cells can
    be selected by clicking on that cell.
    """
    
    __gsignals__ = {
        "selection-changed" : (gobject.SIGNAL_RUN_LAST,
                               gobject.TYPE_NONE,
                               (gobject.TYPE_PYOBJECT,)),
    }
    
    def __init__(self):
        gtk.TreeView.__init__(self)
        self.description = None
        self.set_rules_hint(True)
        self.get_selection().set_mode(gtk.SELECTION_NONE)
        self.selected_columns = list()
        self.selected_rows = list()
        self.connect("button-press-event", self.on_button_pressed)
        
    def _setup_columns(self, rows):
        renderer = gtk.CellRendererText()
        renderer.set_property("background", self.get_style().dark[gtk.STATE_ACTIVE].to_string())
        renderer.set_property("alignment", pango.ALIGN_CENTER)
        renderer.set_property("xalign", 0)
        renderer.set_property("yalign", 0)
        renderer.set_property("width-chars", len(str(len(rows))))
        col = gtk.TreeViewColumn("#", renderer,
                                 text=(len(self.description)*4)+1
                                 )
        col.connect("clicked", self.on_first_header_clicked)
        col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        col.set_resizable(True)
        col.set_min_width(50)
        col.set_data("pressed", False)
        self.append_column(col)
        offset_fg = len(self.description)*2
        offset_bg = len(self.description)*3
        for i in range(len(self.description)):
            item = self.description[i]
            renderer = gtk.CellRendererText()
            renderer.set_property("ellipsize", pango.ELLIPSIZE_END)
            lbl = "%s" % (item[0].replace("_", "__"),)
            col = gtk.TreeViewColumn(lbl, renderer,
                                     markup=i, 
                                     foreground_gdk=offset_fg+i,
                                     background_gdk=offset_bg+i)
            col.connect("clicked", self.on_column_header_clicked)
            col.set_resizable(True)
            col.set_min_width(75)
            col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
            self.append_column(col)
        self.set_headers_clickable(True)
        self.set_fixed_height_mode(True)
        
    def _setup_model(self, rows, description, coding_hint):
        model = GridModel(rows, description, self.get_style(),
                          coding_hint=coding_hint)
        old_model = self.get_model()
        if old_model:
            del old_model
        self.set_model(model)
        
    def _get_popup_for_cell(self, row, col):
        col = self.get_model_index(col)
        model = self.get_model()
        data = model.on_get_value(row, col+len(self.description)-1)
        popup = gtk.Menu()
        if data != None:
            if not isinstance(data, buffer):
                item = gtk.MenuItem(_(u"Copy value to clipboard"))
                item.connect("activate", self.on_copy_value_to_clipboard, data)
                item.show()
                popup.append(item)
                sep = gtk.SeparatorMenuItem()
                sep.show()
                popup.append(sep)
            if isinstance(data, buffer):
                mime = gnomevfs.get_mime_type_for_data(data)
                item = gtk.MenuItem(_(u"Save as..."))
                item.connect("activate", self.on_save_blob, data, mime)
                item.show()
                popup.append(item)
                if mime:
                    apps = gnomevfs.mime_get_all_applications(mime)
                    if apps:
                        item = gtk.MenuItem(_(u"Open with..."))
                        smenu = gtk.Menu()
                        item.set_submenu(smenu)
                        item.show()
                        popup.append(item)
                        for app_info in apps:
                            item = gtk.MenuItem(app_info[1])
                            item.connect("activate", self.on_open_blob, data, app_info)
                            item.show()
                            smenu.append(item)
            else:
                item = gtk.MenuItem(_(u"View value"))
                item.connect("activate", self.on_view_data, data)
                item.show()
                popup.append(item)
        else:
            item = gtk.MenuItem(_(u"This cell contains a 'NULL' value."))
            item.set_sensitive(False)
            item.show()
            popup.append(item)
        return popup
        
    def on_button_pressed(self, treeview, event):
        if event.type == gtk.gdk._2BUTTON_PRESS:
            x = int(event.x)
            y = int(event.y)
            time = event.time
            pthinfo = treeview.get_path_at_pos(x, y)
            if pthinfo is not None:
                path, col, cellx, celly = pthinfo
                if col == self.get_columns()[0]:
                    return
                col = self.get_model_index(col)
                model = self.get_model()
                data = model.on_get_value(path[0], col+len(self.description)-1)
                if data == None:
                    return
                if not isinstance(data, buffer):
                    self.on_view_data(None, data)
        
        elif event.button == 3:
            x = int(event.x)
            y = int(event.y)
            time = event.time
            pthinfo = treeview.get_path_at_pos(x, y)
            if pthinfo is not None:
                path, col, cellx, celly = pthinfo
                if col == self.get_columns()[0]:
                    return
                self.select_cell(path[0], col, not self.cell_is_selected(path[0], col))
                treeview.grab_focus()
                treeview.set_cursor( path, col, 0)
                popup = self._get_popup_for_cell(path[0], col)
                if not popup:
                    return
                popup.popup( None, None, None, event.button, time)
        
        elif event.button == 1:
            x = int(event.x)
            y = int(event.y)
            pthinfo = treeview.get_path_at_pos(x, y)
            if not pthinfo:
                return
            path, col, cellx, celly = pthinfo
            if col == self.get_columns()[0]:
                self.select_row(path[0], not self.row_is_selected(path[0]))
            else:
                self.select_cell(path[0], col, not self.cell_is_selected(path[0], col))
                
        
            
    def on_column_header_clicked(self, column):
        selected = not column in self.get_selected_columns()
        self.select_column(column, selected)
        
    def on_copy_value_to_clipboard(self, menuitem, value):
        display = gtk.gdk.display_manager_get().get_default_display()
        clipboard = gtk.Clipboard(display, "CLIPBOARD")
        clipboard.set_text(str(value))
        
    def on_first_header_clicked(self, column):
        selected = not column.get_data("pressed")
        for xcolumn in self.get_columns()[1:]:
            self.select_column(xcolumn, selected)
        column.set_data("pressed", selected)
        
    def on_open_blob(self, menuitem, data, app_info):
        fname = os.path.expanduser("~/cf-blob.file")
        cmd = "%s %s" % (app_info[2], fname)
        f = open(fname, "w")
        f.write(str(data))
        f.close()
        os.system(cmd)
        
    def on_save_blob(self, menuitem, data, mime):
        dlg = gtk.FileChooserDialog(_(u"Save as..."),
                                    None,
                                    gtk.FILE_CHOOSER_ACTION_SAVE,
                                    (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                     gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        if mime:
            filter = gtk.FileFilter()
            filter.set_name(mime)
            filter.add_mime_type(mime)
            dlg.add_filter(filter)
            dlg.set_filter(filter)
            filter = gtk.FileFilter()
            filter.set_name(_(u"All files"))
            filter.add_pattern("*")
            dlg.add_filter(filter)
        if dlg.run() == gtk.RESPONSE_OK:
            f = open(dlg.get_filename(), "w")
            f.write(str(data))
            f.close()
        dlg.destroy()
        
    def on_view_data(self, menuitem, data):
        dlg = DataViewer(data)
        dlg.run()
        dlg.destroy()
        
    def cell_is_selected(self, row, column):
        """Returns ``True`` is a cell is selected
        
        :Parameter:
            row
                Row number
            column
                ``gtk.TreeViewColumn``
        """
        col = self.get_model_index(column)-1
        return (row, col) in self.get_model().selected_cells
    
    def get_cell_data(self, cell, repr=False):
        """Returns the content of a cell
        
        :Parameter:
            cell
                A row-cell tuple
            repr
                If ``True`` the content of the cell is returned as text (default: ``False``)
        :Returns: Cell content
        """
        model = self.get_model()
        data = model.on_get_value(cell[0], cell[1]+len(self.description))
        if repr:
            data = model._get_markup_for_value(data, strip_length=False, markup=False)
        return data
    
    def get_grid_data(self):
        """Returns all data"""
        return self.get_model().rows
    
    def get_model_index(self, treeview_column):
        """Returns the model index for a column
        
        :Parameter:
            treeview_column
                ``gtk.TreeViewColumn`
        
        .. Note:: 
            This method returns the label column.
        """
        columns = self.get_columns()
        for i in range(len(columns)):
            if columns[i] == treeview_column:
                return i
            
    def get_selected_cells(self):
        """Returns selected cells"""
        return self.get_model().selected_cells
    
    def get_selected_columns(self):
        """Returns selected columns"""
        return self.selected_columns
    
    def get_selected_rows(self):
        """Returns selected rows"""
        return self.selected_rows
        
    def reset(self):
        """Resets the grid"""
        old_model = self.get_model()
        if old_model:
            del old_model
        model = gtk.ListStore(int)
        self.set_model(model)
        while self.get_columns():
            col = self.get_column(0)
            self.remove_column(col)
            
    def row_is_selected(self, row):
        """Returns ``True`` if the row is selected
        
        :Parameter:
            row
                Row number
        """
        return row in self.selected_rows
    
    def select_cell(self, row, column, selected):
        """Selects a cell
        
        :Parameter:
            row
                A row index
            column
                A ``gtk.TreeViewColumn``
            selected
                ``True`` if the cell should be selected
        """
        col = self.get_model_index(column)-1
        path = (row, col)
        model = self.get_model()
        self.unselect_cells()
        if selected:
            self.unselect_rows()
            self.unselect_columns()
        if selected and path not in model.selected_cells:
            model.selected_cells.append(path)
        elif not selected and path in model.selected_cells:
            model.selected_cells.remove(path)
        self.emit("selection-changed", model.selected_cells)
            
    def select_column(self, column, selected):
        """Selects a column
        
        :Parameter:
            column
                a ``gtk.TreeViewColumn``
            selected
                ``True`` if the column should be selected
        """
        if column in self.get_selected_columns() \
        and not selected:
            style = self.get_style().bg[gtk.STATE_NORMAL]
            background_set = False
            self.selected_columns.remove(column)
        elif column not in self.get_selected_columns() \
        and selected:
            style = self.get_style().bg[gtk.STATE_SELECTED]
            background_set = True
            self.unselect_rows()
            self.unselect_cells()
            self.selected_columns.append(column)
        else:
            return
        for renderer in column.get_cell_renderers():
            renderer.set_property("cell-background-gdk", style)
            renderer.set_property("cell-background-set", background_set)
        # rebuild selected_cells
        self.unselect_cells()
        model = self.get_model()
        for i in range(len(model.rows)):
            for column in self.get_selected_columns():
                j = self.get_model_index(column)-1
                model.selected_cells.append((i, j))
        self.queue_draw()
        self.emit("selection-changed", model.selected_cells)
        
    def select_row(self, row, selected):
        """Selects a row
        
        :Parameter:
            row
                the row number
            selected
                ``True`` if the row should be selected
        """
        if selected and not self.row_is_selected(row):
            self.unselect_columns()
            if not self.selected_rows:
                self.unselect_cells()
            self.selected_rows.append(row)
            model = self.get_model()
            for i in range(len(self.get_columns())):
                path = (row, i)
                model.selected_cells.append(path)
        elif not selected and self.row_is_selected(row):
            self.selected_rows.remove(row)
            model = self.get_model()
            for i in range(len(self.get_columns())):
                path = (row, i)
                if path in model.selected_cells:
                    model.selected_cells.remove(path)
        else:
            return
        self.queue_draw()
        self.emit("selection-changed", model.selected_cells)
    
    def set_result(self, rows, description, coding_hint="utf-8"):
        """Sets the result and updates the grid
        
        :Parameter:
            rows
                Sequence of rows
            description
                DB-API2 like description
        """
        self.description = description
        self._setup_columns(rows)
        self._setup_model(rows, description, coding_hint)
        
    def unselect_cells(self):
        """Unselects all cells"""
        model = self.get_model()
        model.selected_cells = []
        
    def unselect_columns(self):
        """Unselects all columns"""
        while self.get_selected_columns():
            column = self.get_selected_columns()[0]
            self.select_column(column, False)
            
    def unselect_rows(self):
        """Unselect all rows"""
        while self.get_selected_rows():
            row = self.get_selected_rows()[0]
            self.select_row(row, False)
        
class GridModel(gtk.GenericTreeModel):
    """Data grid model
    
    The model stores it's data in a plain Python list. It provides
    three virtual columns for a displayed version of a value (limited
    to ``GRID_LABEL_MAX_LENGTH`` characters to increase perfomance),
    a foreground and a background color for selected cells.
    
    This class re-uses some code of the `Nicotine`_ project (`FastListModel`_)
    found via Google's Code Search.
    
    .. _Nicotine: http://nicotine-plus.sourceforge.net/
    .. _FastListModel: http://www.google.com/codesearch?hl=de&q=+lang:python+GenericTreeModel+show:VRnMlwyOXFM:6NW9oRiVVfg:ANlgLtp-rX8&sa=N&cd=20&ct=rc&cs_p=http://ftp.tr.freebsd.org/pub/FreeBSD/distfiles/nicotine%2B-1.2.6.tar.bz2&cs_f=nicotine%2B-1.2.6/pynicotine/gtkgui/utils.py#first
    """
    
    def __init__(self, rows, description, style, coding_hint="utf-8"):
        """
        The constructor takes three arguments:
        
        :Parameter:
            rows
                Data as a plain python list of rows
            description
                DB-API2 like description
            style
                ``gtk.Style`` instance
        """
        gtk.GenericTreeModel.__init__(self)
        self.rows = rows
        self.description = description
        self.style = style
        self.coding_hint = coding_hint
        self.selected_cells = list()
        
    def _get_markup_for_value(self, value, strip_length=True, markup=True):
        style = self.style
        if value == None: 
            if markup:
                value = '<span foreground="%s">&lt;NULL&gt;</span>' % style.dark[gtk.STATE_PRELIGHT].to_string()
            else:
                value = 'null'
        elif isinstance(value, buffer):
            if markup:
                value = '<span foreground="%s">&lt;LOB&gt;</span>' % style.dark[gtk.STATE_PRELIGHT].to_string()
            else:
                value = str(buffer)
        else:
            if isinstance(value, str):
                value = unicode(value, self.coding_hint)
            elif isinstance(value, unicode):
                pass
            else:
                value = unicode(value)
            value = value.splitlines()
            if value:
                if strip_length:
                    value = value[0][:GRID_LABEL_MAX_LENGTH]
                else:
                    value = "\n".join(value)
            else:
                value = ""
            value = gobject.markup_escape_text(value)
        return value

    def on_get_flags(self):
        '''returns the GtkTreeModelFlags for this particular type of model'''
        return gtk.TREE_MODEL_LIST_ONLY

    def on_get_n_columns(self):
        '''returns the number of columns in the model'''
        return len(self.description)*4+1

    def on_get_column_type(self, index):
        '''returns the type of a column in the model'''
        length = len(self.description)
        range_label = range(length)
        range_data = range(length, length*2)
        range_fg = range(length*2, length*3)
        range_bg = range(length*3, length*4)
        if index in range_label:
            return str
        elif index in range_data:
            return gobject.TYPE_PYOBJECT
        elif index == self.on_get_n_columns():
            return int
        elif (index in range_fg) or (index in range_bg):
            return gtk.gdk.Color
        else:
            raise RuntimeError, "Unexpected index"

    def on_get_path(self, iter):
        '''returns the tree path (a tuple of indices at the various
        levels) for a particular node.'''
        return (iter,)

    def on_get_iter(self, path):
        '''returns the node corresponding to the given path.  In our
        case, the node is the path'''
        if path[0] < len(self.rows):
            return path[0]
        else:
            return None

    def on_get_value(self, iter, column):
        '''returns the value stored in a particular column for the node'''
        length = len(self.description)
        range_label = range(length)
        range_data = range(length, length*2)
        range_fg = range(length*2, length*3)
        range_bg = range(length*3, length*4)
        if column in range_label:
            raw = self.rows[iter][column]
            markup = self._get_markup_for_value(raw)
            return markup
        elif column in range_data:
            return self.rows[iter][column-len(self.description)]
        elif column == self.on_get_n_columns():
            return iter+1
        elif column in range_fg:
            data_column = column-length*2
            if (iter, data_column) in self.selected_cells:
                return self.style.fg[gtk.STATE_SELECTED]
            else:
                return None
        elif column in range_bg:
            data_column = column-length*3
            if (iter, data_column) in self.selected_cells:
                return self.style.bg[gtk.STATE_SELECTED]
            else:
                return None
        else:
            raise RuntimeError, "Unexpected index %r" % column 

    def on_iter_next(self, iter):
        '''returns the next node at this level of the tree'''
        if iter + 1 < len(self.rows):
            return iter + 1
        else:
            return None

    def on_iter_children(self, iter):
        '''returns the first child of this node'''
        return 0

    def on_iter_has_child(self, iter):
        '''returns true if this node has children'''
        return False

    def on_iter_n_children(self, iter):
        '''returns the number of children of this node'''
        return len(self.rows)

    def on_iter_nth_child(self, iter, n):
        '''returns the nth child of this node'''
        return n

    def on_iter_parent(self, iter):
        '''returns the parent of this node'''
        return None
    
    
class DataViewer(gtk.Dialog):
    """Dialog to display a value"""
    
    def __init__(self, data):
        """
        The constructor of this class takes 1 argument:
        
        :Parameter:
            data
                A Python value to display
        """
        gtk.Dialog.__init__(self, _(u"Data"), None,
                            gtk.DIALOG_DESTROY_WITH_PARENT,
                            (gtk.STOCK_CLOSE, gtk.RESPONSE_OK))
        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        sw.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        self.vbox.pack_start(sw, True, True)
        tv = gtk.TextView()
        tv.set_editable(False)
        tv.set_wrap_mode(gtk.WRAP_WORD_CHAR)
        tv.get_buffer().set_text(str(data))
        sw.add(tv)
        sw.show_all()
        self.resize(650, 550)
        



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: __init__.py 71 2008-03-10 20:57:35Z albrecht.andi $

"""User interface"""

import gobject
import gtk

from os.path import abspath, join, dirname

import cf

class GladeWidget(gobject.GObject):
    """Helper for Glade widgets
    
    This class is a helper for widgets defined in Glade files. It
    shouldn't be used directly.
    
    Upon intialization the two methods ``_setup_widget()`` and 
    ``_setup_connections()`` are called.
    
    Read the source of one of the subclasses to get an idea how
    this class works.
    
    Instance attributes
    ===================
    
        :app: `CFApplication`_ instance
        :xml: `gtk.glade.XML`_ instance
        :widget: The *real* GTK widget
        
    .. _CFApplication: cf.app.CFApplication.html
    .. _gtk.glade.XML: http://pygtk.org/docs/pygtk/class-gladexml.html
    """
    
    def __init__(self, app, xml, widget_name,
                 signal_autoconnect=True,
                 cb_provider=None):
        """
        This constructor takes up to five arguments:
        
        :Parameter:
            app
                `CFApplication`_ instance
            xml
                Glade source
            widget_name
                Name of the widget to fetch
            signal_autoconnect
                Auto-connect signals (optional, default ``True``)
            cb_provider
                Argument for ``signal_autoconnect`` (optional)
                
        If ``xml`` is a string it is assumed that it is the name of
        a Glade file located in the data directory. If it's a `gtk.glade.XML`_
        instance, this instance is taken as the widget source.
        
        If ``signal_autoconnect`` is ``True`` the ``signal_autoconnect()`` method
        of the `gtk.glade.XML`_ instance is called upon initialization.
        If the optional ``cb_provider`` argument is given, it will be used as
        the parameter for ``signal_autoconnect()``. Otherwise 
        ``signal_autoconnect(self)`` is called.
        
        .. _CFApplication: cf.app.CFApplication.html
        .. _gtk.glade.XML: http://pygtk.org/docs/pygtk/class-gladexml.html 
        """
        self.__gobject_init__()
        self.app = app
        if isinstance(xml, basestring):
            if not xml.endswith(".glade"):
                xml += ".glade"
            self.xml = gtk.glade.XML(join(cf.DATA_DIR, xml), widget_name)
        else:
            self.xml = xml
        if signal_autoconnect:
            if not cb_provider:
                cb_provider = self
            self.xml.signal_autoconnect(cb_provider)
        self.widget = self.xml.get_widget(widget_name)
        for item in dir(self.widget):
            if item.startswith("_") \
            or hasattr(self, item): continue
            setattr(self, item, getattr(self.widget, item))
        self.widget.set_data("glade-widget", self)
        self._setup_widget()
        self._setup_connections()
        
    def _setup_widget(self):
        """Called when the object is created."""
        
    def _setup_connections(self):
        """Called when the object is created."""
        


#!/usr/bin/python
'''
Provides IPython console widget.

@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made 
available under the terms of the BSD which accompanies this distribution, and 
is available at U{http://www.opensource.org/licenses/bsd-license.php}

http://ipython.scipy.org/moin/Cookbook/EmbeddingInGTK
'''

import gtk, gobject
import re
import sys
import os
import pango
from StringIO import StringIO
import thread

try:
  import IPython
except ImportError:
  IPython = None

class IterableIPShell:
  '''
  Create an IPython instance. Does not start a blocking event loop,
  instead allow single iterations. This allows embedding in GTK+ 
  without blockage.

  @ivar IP: IPython instance.
  @type IP: IPython.iplib.InteractiveShell
  @ivar iter_more: Indicates if the line executed was a complete command,
  or we should wait for more.
  @type iter_more: integer
  @ivar history_level: The place in history where we currently are 
  when pressing up/down.
  @type history_level: integer
  @ivar complete_sep: Seperation delimeters for completion function.
  @type complete_sep: _sre.SRE_Pattern
  '''
  def __init__(self,argv=[],user_ns=None,user_global_ns=None, 
               cin=None, cout=None,cerr=None, input_func=None):
    '''
    
    
    @param argv: Command line options for IPython
    @type argv: list
    @param user_ns: User namespace.
    @type user_ns: dictionary
    @param user_global_ns: User global namespace.
    @type user_global_ns: dictionary.
    @param cin: Console standard input.
    @type cin: IO stream
    @param cout: Console standard output.
    @type cout: IO stream 
    @param cerr: Console standard error.
    @type cerr: IO stream
    @param input_func: Replacement for builtin raw_input()
    @type input_func: function
    '''
    if input_func:
      IPython.iplib.raw_input_original = input_func
    if cin:
      IPython.Shell.Term.cin = cin
    if cout:
      IPython.Shell.Term.cout = cout
    if cerr:
      IPython.Shell.Term.cerr = cerr

    # This is to get rid of the blockage that accurs during 
    # IPython.Shell.InteractiveShell.user_setup()
    IPython.iplib.raw_input = lambda x: None

    self.term = IPython.genutils.IOTerm(cin=cin, cout=cout, cerr=cerr)
    os.environ['TERM'] = 'dumb'
    excepthook = sys.excepthook 
    self.IP = IPython.Shell.make_IPython(
      argv,user_ns=user_ns,
      user_global_ns=user_global_ns,
      embedded=True,
      shell_class=IPython.Shell.InteractiveShell)
    self.IP.system = lambda cmd: self.shell(self.IP.var_expand(cmd),
                                            header='IPython system call: ',
                                            verbose=self.IP.rc.system_verbose)
    sys.excepthook = excepthook
    self.iter_more = 0
    self.history_level = 0
    self.complete_sep =  re.compile('[\s\{\}\[\]\(\)]')

  def execute(self):
    '''
    Executes the current line provided by the shell object.
    '''
    self.history_level = 0
    orig_stdout = sys.stdout
    sys.stdout = IPython.Shell.Term.cout
    try:
      line = self.IP.raw_input(None, self.iter_more)
      if self.IP.autoindent:
        self.IP.readline_startup_hook(None)
    except KeyboardInterrupt:
      self.IP.write('\nKeyboardInterrupt\n')
      self.IP.resetbuffer()
      # keep cache in sync with the prompt counter:
      self.IP.outputcache.prompt_count -= 1
        
      if self.IP.autoindent:
        self.IP.indent_current_nsp = 0
      self.iter_more = 0
    except:
      self.IP.showtraceback()
    else:
      self.iter_more = self.IP.push(line)
      if (self.IP.SyntaxTB.last_syntax_error and
          self.IP.rc.autoedit_syntax):
        self.IP.edit_syntax_error()
    if self.iter_more:
      self.prompt = str(self.IP.outputcache.prompt2).strip()
      if self.IP.autoindent:
        self.IP.readline_startup_hook(self.IP.pre_readline)
    else:
      self.prompt = str(self.IP.outputcache.prompt1).strip()
    sys.stdout = orig_stdout

  def historyBack(self):
    '''
    Provides one history command back.
    
    @return: The command string.
    @rtype: string
    '''
    self.history_level -= 1
    return self._getHistory()
  
  def historyForward(self):
    '''
    Provides one history command forward.
    
    @return: The command string.
    @rtype: string
    '''
    self.history_level += 1
    return self._getHistory()
  
  def _getHistory(self):
    '''
    Get's the command string of the current history level.
    
    @return: Historic command string.
    @rtype: string
    '''
    try:
      rv = self.IP.user_ns['In'][self.history_level].strip('\n')
    except IndexError:
      self.history_level = 0
      rv = ''
    return rv

  def updateNamespace(self, ns_dict):
    '''
    Add the current dictionary to the shell namespace.
    
    @param ns_dict: A dictionary of symbol-values.
    @type ns_dict: dictionary
    '''
    self.IP.user_ns.update(ns_dict)

  def complete(self, line):
    '''
    Returns an auto completed line and/or posibilities for completion.
    
    @param line: Given line so far.
    @type line: string
    
    @return: Line completed as for as possible, 
    and possible further completions.
    @rtype: tuple
    '''
    split_line = self.complete_sep.split(line)
    possibilities = self.IP.complete(split_line[-1])
    if possibilities:
      def _commonPrefix(str1, str2):
        '''
        Reduction function. returns common prefix of two given strings.
        
        @param str1: First string.
        @type str1: string
        @param str2: Second string
        @type str2: string
        
        @return: Common prefix to both strings.
        @rtype: string
        '''
        for i in range(len(str1)):
          if not str2.startswith(str1[:i+1]):
            return str1[:i]
        return str1
      common_prefix = reduce(_commonPrefix, possibilities)
      completed = line[:-len(split_line[-1])]+common_prefix
    else:
      completed = line
    return completed, possibilities
  

  def shell(self, cmd,verbose=0,debug=0,header=''):
    '''
    Replacement method to allow shell commands without them blocking.
    
    @param cmd: Shell command to execute.
    @type cmd: string
    @param verbose: Verbosity
    @type verbose: integer
    @param debug: Debug level
    @type debug: integer
    @param header: Header to be printed before output
    @type header: string
    '''
    stat = 0
    if verbose or debug: print header+cmd
    # flush stdout so we don't mangle python's buffering
    if not debug:
      input, output = os.popen4(cmd)
      print output.read()
      output.close()
      input.close()

class ConsoleView(gtk.TextView):
  '''
  Specialized text view for console-like workflow.

  @cvar ANSI_COLORS: Mapping of terminal colors to X11 names.
  @type ANSI_COLORS: dictionary

  @ivar text_buffer: Widget's text buffer.
  @type text_buffer: gtk.TextBuffer
  @ivar color_pat: Regex of terminal color pattern
  @type color_pat: _sre.SRE_Pattern
  @ivar mark: Scroll mark for automatic scrolling on input.
  @type mark: gtk.TextMark
  @ivar line_start: Start of command line mark.
  @type line_start: gtk.TextMark
  '''
  ANSI_COLORS =  {'0;30': 'Black',     '0;31': 'Red',
                  '0;32': 'Green',     '0;33': 'Brown',
                  '0;34': 'Blue',      '0;35': 'Purple',
                  '0;36': 'Cyan',      '0;37': 'LightGray',
                  '1;30': 'DarkGray',  '1;31': 'DarkRed',
                  '1;32': 'SeaGreen',  '1;33': 'Yellow',
                  '1;34': 'LightBlue', '1;35': 'MediumPurple',
                  '1;36': 'LightCyan', '1;37': 'White'}

  def __init__(self):
    '''
    Initialize console view.
    '''
    gtk.TextView.__init__(self)
    self.modify_font(pango.FontDescription('Mono'))
    self.set_cursor_visible(True)
    self.text_buffer = self.get_buffer()
    self.mark = self.text_buffer.create_mark('scroll_mark', 
                                             self.text_buffer.get_end_iter(),
                                             False)
    for code in self.ANSI_COLORS:
      self.text_buffer.create_tag(code, 
                                  foreground=self.ANSI_COLORS[code], 
                                  weight=700)
    self.text_buffer.create_tag('0')
    self.text_buffer.create_tag('notouch', editable=False)
    self.color_pat = re.compile('\x01?\x1b\[(.*?)m\x02?')
    self.line_start = \
        self.text_buffer.create_mark('line_start', 
                                     self.text_buffer.get_end_iter(), True)
    self.connect('key-press-event', self.onKeyPress)
    
  def write(self, text, editable=False):
    gobject.idle_add(self._write, text, editable)

  def _write(self, text, editable=False):
    '''
    Write given text to buffer.
    
    @param text: Text to append.
    @type text: string
    @param editable: If true, added text is editable.
    @type editable: boolean
    '''
    segments = self.color_pat.split(text)
    segment = segments.pop(0)
    start_mark = self.text_buffer.create_mark(None, 
                                              self.text_buffer.get_end_iter(), 
                                              True)
    self.text_buffer.insert(self.text_buffer.get_end_iter(), segment)

    if segments:
      ansi_tags = self.color_pat.findall(text)
      for tag in ansi_tags:
        i = segments.index(tag)
        self.text_buffer.insert_with_tags_by_name(self.text_buffer.get_end_iter(),
                                             segments[i+1], tag)
        segments.pop(i)
    if not editable:
      self.text_buffer.apply_tag_by_name('notouch',
                                         self.text_buffer.get_iter_at_mark(start_mark),
                                         self.text_buffer.get_end_iter())
    self.text_buffer.delete_mark(start_mark)
    self.scroll_mark_onscreen(self.mark)


  def showPrompt(self, prompt):
    gobject.idle_add(self._showPrompt, prompt)

  def _showPrompt(self, prompt):
    '''
    Prints prompt at start of line.
    
    @param prompt: Prompt to print.
    @type prompt: string
    '''
    self._write(prompt)
    self.text_buffer.move_mark(self.line_start,
                               self.text_buffer.get_end_iter())

  def changeLine(self, text):
    gobject.idle_add(self._changeLine, text)

  def _changeLine(self, text):
    '''
    Replace currently entered command line with given text.
    
    @param text: Text to use as replacement.
    @type text: string
    '''
    iter = self.text_buffer.get_iter_at_mark(self.line_start)
    iter.forward_to_line_end()
    self.text_buffer.delete(self.text_buffer.get_iter_at_mark(self.line_start), iter)
    self._write(text, True)

  def getCurrentLine(self):
    '''
    Get text in current command line.
    
    @return: Text of current command line.
    @rtype: string
    '''
    rv = self.text_buffer.get_slice(
      self.text_buffer.get_iter_at_mark(self.line_start),
      self.text_buffer.get_end_iter(), False)
    return rv

  def showReturned(self, text):
    gobject.idle_add(self._showReturned, text)

  def _showReturned(self, text):
    '''
    Show returned text from last command and print new prompt.
    
    @param text: Text to show.
    @type text: string
    '''
    iter = self.text_buffer.get_iter_at_mark(self.line_start)
    iter.forward_to_line_end()
    self.text_buffer.apply_tag_by_name(
      'notouch', 
      self.text_buffer.get_iter_at_mark(self.line_start),
      iter)
    self._write('\n'+text)
    if text:
      self._write('\n')
    self._showPrompt(self.prompt)
    self.text_buffer.move_mark(self.line_start,self.text_buffer.get_end_iter())
    self.text_buffer.place_cursor(self.text_buffer.get_end_iter())

  def onKeyPress(self, widget, event):
    '''
    Key press callback used for correcting behavior for console-like 
    interfaces. For example 'home' should go to prompt, not to begining of
    line.
    
    @param widget: Widget that key press accored in.
    @type widget: gtk.Widget
    @param event: Event object
    @type event: gtk.gdk.Event
    
    @return: Return True if event should not trickle.
    @rtype: boolean
    '''
    insert_mark = self.text_buffer.get_insert()
    insert_iter = self.text_buffer.get_iter_at_mark(insert_mark)
    selection_mark = self.text_buffer.get_selection_bound()
    selection_iter = self.text_buffer.get_iter_at_mark(selection_mark)
    start_iter = self.text_buffer.get_iter_at_mark(self.line_start)
    if event.keyval == gtk.keysyms.Home:
      if event.state == 0: 
        self.text_buffer.place_cursor(start_iter)
        return True
      elif event.state == gtk.gdk.SHIFT_MASK:
        self.text_buffer.move_mark(insert_mark, start_iter)
        return True
    elif event.keyval == gtk.keysyms.Left:
      insert_iter.backward_cursor_position()
      if not insert_iter.editable(True):
        return True
    elif not event.string:
      pass
    elif start_iter.compare(insert_iter) <= 0 and \
          start_iter.compare(selection_iter) <= 0:
      pass
    elif start_iter.compare(insert_iter) > 0 and \
          start_iter.compare(selection_iter) > 0:
      self.text_buffer.place_cursor(start_iter)
    elif insert_iter.compare(selection_iter) < 0:
      self.text_buffer.move_mark(insert_mark, start_iter)
    elif insert_iter.compare(selection_iter) > 0:
      self.text_buffer.move_mark(selection_mark, start_iter)             

    return self.onKeyPressExtend(event)

  def onKeyPressExtend(self, event):
    '''
    For some reason we can't extend onKeyPress directly (bug #500900).
    '''
    pass

class IPythonView(ConsoleView, IterableIPShell):
  '''
  Sub-class of both modified IPython shell and L{ConsoleView} this makes
  a GTK+ IPython console.
  '''
  def __init__(self):
    '''
    Initialize. Redirect I/O to console.
    '''
    ConsoleView.__init__(self)
    self.cout = StringIO()
    IterableIPShell.__init__(self, cout=self.cout,cerr=self.cout, 
                             input_func=self.raw_input)
#    self.connect('key_press_event', self.keyPress)
    self.execute()
    self.cout.truncate(0)
    self.showPrompt(self.prompt)
    self.interrupt = False

  def raw_input(self, prompt=''):
    '''
    Custom raw_input() replacement. Get's current line from console buffer.
    
    @param prompt: Prompt to print. Here for compatability as replacement.
    @type prompt: string
    
    @return: The current command line text.
    @rtype: string
    '''
    if self.interrupt:
      self.interrupt = False
      raise KeyboardInterrupt
    return self.getCurrentLine()

  def onKeyPressExtend(self, event):
    '''
    Key press callback with plenty of shell goodness, like history, 
    autocompletions, etc.
    
    @param widget: Widget that key press occured in.
    @type widget: gtk.Widget
    @param event: Event object.
    @type event: gtk.gdk.Event
    
    @return: True if event should not trickle.
    @rtype: boolean
    '''
    if event.state & gtk.gdk.CONTROL_MASK and event.keyval == 99:
      self.interrupt = True
      self._processLine()
      return True
    elif event.keyval == gtk.keysyms.Return:
      self._processLine()
      return True
    elif event.keyval == gtk.keysyms.Up:
      self.changeLine(self.historyBack())
      return True
    elif event.keyval == gtk.keysyms.Down:
      self.changeLine(self.historyForward())
      return True
    elif event.keyval == gtk.keysyms.Tab:
      if not self.getCurrentLine().strip():
        return False
      completed, possibilities = self.complete(self.getCurrentLine())
      if len(possibilities) > 1:
        slice = self.getCurrentLine()
        self.write('\n')
        for symbol in possibilities:
          self.write(symbol+'\n')
        self.showPrompt(self.prompt)
      self.changeLine(completed or slice)
      return True

  def _processLine(self):
    '''
    Process current command line.
    '''
    self.history_pos = 0
    self.execute()
    rv = self.cout.getvalue()
    if rv: rv = rv.strip('\n')
    self.showReturned(rv)
    self.cout.truncate(0)
    
